@@ -5,7 +5,7 @@
  * spread across multiple lines.
  */
 
-/* Copyright (c) 2018-2020 MJ Rutter 
+/* Copyright (c) 2018-2021 MJ Rutter 
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -31,6 +31,7 @@
 #include "c2xsf.h"
 
 #define LINE_SIZE 132  /* Thus says abinit */
+#define SLINE_SIZE 300 /* For Siesta, N.B. SLINE_SIZE >= LINE_SIZE */
 
 static int ab_len_read(char *line, int off, double *data, int n,
 		       FILE* infile);
@@ -40,7 +41,7 @@ static int ab_float_read(char *line, int off, double *data, int n,
 			  FILE* infile);
 static int ab_int_read(char *line, int off, int *data, int n,
 		       FILE* infile);
-static int ab_string_read(char *line, int off, char **str);
+static int ab_string_read(char *line, int off, char **str, FILE* infile);
 static int ab_readline(char *buffer, char **p, FILE* infile);
 
 /* from cif_read */
@@ -60,7 +61,7 @@ void abinit_in_read(FILE* in, struct unit_cell *c, struct contents *m,
   
   double *acell,*rprim,*scalecart,*znucl,*xcart,*xred,*kpt,*wtk,*shiftk;
   double *ecut,*toldfe,*spinat,*tnons,*kptnrm,*zion;
-  double x,xa,xc,angles[3],mat[3][3];
+  double x,xa,xc,angles[3],mat[3][3],abc[6];
   char line[LINE_SIZE+1];
   char *ptr,*p2;
   int i,j,jj,success,ignore;
@@ -74,6 +75,7 @@ void abinit_in_read(FILE* in, struct unit_cell *c, struct contents *m,
    * first pass */
   char *pass2[]={"typat","xangst","xcart","xred","znucl","zion","spinat",
 		 "kpt","wtk","shiftk","symrel","tnons",""};
+  struct unit_cell dummy_cell;
   
   acell=rprim=scalecart=znucl=xcart=xred=kpt=wtk=shiftk=NULL;
   ecut=toldfe=spinat=tnons=kptnrm=zion=NULL;
@@ -83,11 +85,13 @@ void abinit_in_read(FILE* in, struct unit_cell *c, struct contents *m,
   *ptr=0;
 
   files=malloc(sizeof(struct infiles));
-  if (!files) error_exit("Malloc error for files in abinit_in_read");
+  if (!files) error_exit("malloc error for files in abinit_in_read");
   files->f=in;
   files->next=files->last=NULL;
   files->name=NULL;
   files->line=0;
+  files->ret=NULL;
+  files->count=0;
   infile=files->f;
   dir=dict_get(m->dict,"in_dir");
   
@@ -117,6 +121,7 @@ void abinit_in_read(FILE* in, struct unit_cell *c, struct contents *m,
 
     if (!tokenmatch(&ptr,"&control")){
       rewind(infile);
+      free(files);
       qe_read(infile,c,m,k,s,e);
       return;
     }
@@ -135,18 +140,26 @@ void abinit_in_read(FILE* in, struct unit_cell *c, struct contents *m,
     }
     
     if (!tokenmatch(&ptr,"include")){
-      ptr+=ab_string_read(ptr,0,&p2);
+      ptr+=ab_string_read(ptr,0,&p2,infile);
       include_file(&files,dir,p2);
       if (debug>1) fprintf(stderr,"Now reading %s\n",ptr);
       free(p2);
       infile=files->f;
     }
     else if (!tokenmatch(&ptr,"pp_dirpath")){
-      ptr+=ab_string_read(ptr,0,&p2);
+      ptr+=ab_string_read(ptr,0,&p2,infile);
       dict_add(m->dict,"Abinit_pp_dirpath",p2);
     }
+    else if (!tokenmatch(&ptr,"output_file")){
+      ptr+=ab_string_read(ptr,0,&p2,infile);
+      dict_add(m->dict,"Abinit_output_file",p2);
+    }
+    else if (!tokenmatch(&ptr,"tmpdata_prefix")){
+      ptr+=ab_string_read(ptr,0,&p2,infile);
+      dict_add(m->dict,"Abinit_tmpdata_prefix",p2);
+    }
     else if (!tokenmatch(&ptr,"pseudos")){
-      ptr+=ab_string_read(ptr,0,&p2);
+      ptr+=ab_string_read(ptr,0,&p2,infile);
       dict_add(m->dict,"Abinit_pseudos",p2);
     }
     else if (!tokenmatch(&ptr,"acell")){
@@ -191,23 +204,10 @@ void abinit_in_read(FILE* in, struct unit_cell *c, struct contents *m,
 	}
       }
       else{
-	for(i=0;i<3;i++)angles[i]*=M_PI/180;
-	rprim[0]=1;
-	rprim[1]=0;
-	rprim[2]=0;
-	rprim[3]=cos(angles[2]);
-	rprim[4]=sin(angles[2]);
-	rprim[5]=0;
-	rprim[6]=cos(angles[1]);
-	x=cos(angles[0])-rprim[0]*rprim[3];
-	rprim[7]=0;
-	if (fabs(rprim[7])>1e-20){
-	  if (fabs(rprim[4])>1e-30)
-	    rprim[7]=x/rprim[4];
-	  else
-	    error_exit("Unable to expand angdeg to rprim");
-	}
-	rprim[8]=sqrt(1-rprim[6]*rprim[6]-rprim[7]*rprim[7]);
+	abc[0]=abc[1]=abc[2]=1;
+	for(i=0;i<3;i++) abc[3+i]=angles[i];
+	dummy_cell.basis=(double(*)[3])rprim;
+	abc2cart(abc,&dummy_cell);
       }	  
     }
     else if (!tokenmatch(&ptr,"ndtset")){
@@ -344,7 +344,7 @@ void abinit_in_read(FILE* in, struct unit_cell *c, struct contents *m,
     }
 
     if (!tokenmatch(&ptr,"include")){
-      ptr+=ab_string_read(ptr,0,&p2);
+      ptr+=ab_string_read(ptr,0,&p2,infile);
       include_file(&files,dir,p2);
       if (debug>1) fprintf(stderr,"Now reading %s\n",ptr);
       free(p2);
@@ -425,7 +425,7 @@ void abinit_in_read(FILE* in, struct unit_cell *c, struct contents *m,
 
   /* basis */
   if (!(c->basis=malloc(9*sizeof(double))))
-    error_exit("Malloc error for basis");
+    error_exit("malloc error for basis");
 
   if (rprim)
     for(i=0;i<3;i++)
@@ -511,9 +511,15 @@ void abinit_in_read(FILE* in, struct unit_cell *c, struct contents *m,
     if ((*nsppol==2)||((*nsppol==1)&&(nspden)&&(*nspden==2))){
       for(i=0;i<m->n;i++)
 	m->atoms[i].spin=spinat[3*i+2];
-      free(spinat);
+    }
+    else if ((nspden)&&(*nspden==4)){
+      for(i=0;i<m->n;i++)
+	for(j=0;j<3;j++)
+	  m->atoms[i].vspin[j]=spinat[3*i+j];
     }
     else fprintf(stderr,"Discarding spins\n");
+    free(spinat);
+    spinat=NULL;
   }
 
   /* symmetry */
@@ -555,7 +561,7 @@ void abinit_in_read(FILE* in, struct unit_cell *c, struct contents *m,
     if (!symrel) error_exit("nsym>1 but symrel not found");
     s->n=*nsym;
     s->ops=malloc(s->n*sizeof(struct sym_op));
-    if (!s->ops) error_exit("Malloc error for symmetry ops");
+    if (!s->ops) error_exit("malloc error for symmetry ops");
     for(i=0;i<s->n;i++){
       for(j=0;j<3;j++)
 	for(jj=0;jj<3;jj++)
@@ -564,8 +570,11 @@ void abinit_in_read(FILE* in, struct unit_cell *c, struct contents *m,
       if (tnons){
 	if ((tnons[3*i]!=0)||(tnons[3*i+1]!=0)||(tnons[3*i+2]!=0)){
 	  s->ops[i].tr=malloc(3*sizeof(double));
+	  if (!s->ops[i].tr) error_exit("malloc error for sym disp");
 	  for(j=0;j<3;j++)
-	    s->ops[i].tr[j]=tnons[3*i+j];
+	    s->ops[i].tr[j]=tnons[3*i]*c->basis[0][j]+
+	                    tnons[3*i+1]*c->basis[1][j]+
+                            tnons[3*i+2]*c->basis[2][j];
 	}
       }
     }
@@ -580,6 +589,7 @@ void abinit_in_read(FILE* in, struct unit_cell *c, struct contents *m,
   if ((kptopt)&&(*kptopt==0)){
     if (!nkpt){
       nkpt=malloc(sizeof(int));
+      if (!nkpt) error_exit("malloc error");
       *nkpt=1;
     }
   }
@@ -653,6 +663,8 @@ void abinit_in_read(FILE* in, struct unit_cell *c, struct contents *m,
     e->etol=*toldfe/m->n;
     free(toldfe);
   }
+
+  free(files);
   
 }
 
@@ -662,11 +674,18 @@ static int ab_len_read(char *line, int off, double *data, int n,
   off=ab_float_read(line,off,data,n,infile);
   /* Do we have some units specified? */
   while (line[off]==' ') off++;
-  if (!strncasecmp(line+off,"angstr",6)){
-    off+=6;
-    while(isalpha(line[off])) off++;
+  if (!strncasecmp(line+off,"ang",3)){ /* "ang", "angstr" and "angstrom" */
+    off+=3;                            /* accepted by Abinit */
+    while(isalpha(line[off])) off++;   /* we accept ang.* */
     return off;
   }
+  if (!strncasecmp(line+off,"nm",2)){
+    off+=2;
+    for(i=0;i<n;i++)
+      data[i]*=10;
+    return off;
+  }
+  if (!strncasecmp(line+off,"bohr",4)) off+=4;
   /* Default for lengths is Bohrs */
   for(i=0;i<n;i++)
     data[i]*=BOHR;
@@ -685,11 +704,19 @@ static int ab_energy_read(char *line, int off, double *data, int n,
     off+=2;
     return off;
   }
-  else if (!strncasecmp(line+off,"Ry",2)){
-    off+=2;
+  if (!strncasecmp(line+off,"meV",3)){
+    off+=3;
+    for(i=0;i<n;i++)
+      data[i]*=0.001;
+    return off;
+  }
+  else if (!strncasecmp(line+off,"Ry",2)){ /* "Ry", "Rydberg" and "Rydbergs" */
+    off+=2;                                /* accepted by Abinit */
+    while(isalpha(line[off])) off++;       /* we accept ry.* */
     for(i=0;i<n;i++)
       data[i]*=0.5;  /* convert to Hartrees, then to eV below */
   }
+  if (!strncasecmp(line+off,"hartree",7)) off+=7;
   /* Default for energy is Hartrees */
   for(i=0;i<n;i++)
     data[i]*=H_eV;
@@ -813,8 +840,11 @@ static int ab_int_read(char *line, int off, int *data, int n,
   return off;
 }
 
-static int ab_string_read(char *line, int off, char **str){
-  char *p1,*p2,*s;
+/* sets s2 to a malloc'd pointer to string read, and returns new
+ * offset in line. Deals with concatenation
+ */
+static int ab_string_read(char *line, int off, char **str, FILE *infile){
+  char *p1,*p2,*s,*s2;
 
   p1=line+off;
   while(isspace(*p1)) p1++;
@@ -823,11 +853,27 @@ static int ab_string_read(char *line, int off, char **str){
   p2=strchr(p1,'"');
   if (!p2) error_exit("Failed to find closing quote for string");
   s=malloc((p2-p1)+1);
-  if (!s) error_exit("Malloc error in string_read");
+  if (!s) error_exit("malloc error in string_read");
   memcpy(s,p1,p2-p1);
   s[p2-p1]=0;
+  p2++;
+  while(isspace(*p2)) p2++;
+  if ((*p2=='/')&&(*(p2+1)=='/')){
+    p2+=2;
+    while(isspace(*p2)) p2++;
+    if (*p2==0){
+      ab_readline(line,&p2,infile);
+      while(isspace(*p2)) p2++;
+    }
+    if (*p2!='"') error_exit("Failed to find initial quote in 2nd string");
+    ab_string_read(line,p2-line,&s2,infile);
+    s=realloc(s,strlen(s)+strlen(s2)+1);
+    if (!s) error_exit("realloc failure in ab_string_read");
+    strcat(s,s2);
+    free(s2);
+  }
   *str=s;
-  return(p2+1-line);
+  return(p2-line);
 }
 
 static int ab_readline(char *buffer, char **p, FILE* infile){
@@ -859,21 +905,35 @@ static int ab_readline(char *buffer, char **p, FILE* infile){
 
 void abinit_eig_read(FILE* infile, struct unit_cell *c, struct contents *m,
 		    struct kpts *k, struct symmetry *s, struct es *e){
-  double scale,dtmp,coord[3];
-  int itmp,i,j,isp,dummy,*gamma,fft[3],fileform;
-  char buffer[LINE_SIZE+1],*ptr,*ptr2,*newfile;
+  double scale,dtmp,dtmp2,coord[3];
+  int itmp,i,j,isp,dummy,*gamma,fft[3],fileform,lineno,no,nspins,nk,siesta;
+  char buffer[SLINE_SIZE+1],*ptr,*ptr2,*newfile;
   FILE *f;
   struct kpts old_k;
   
   /* read header */
 
   scale=0;
+  lineno=0;
+  siesta=0;
   while(1){
     if (!fgets(buffer,LINE_SIZE,infile)) error_exit("Unexpected EOF");
+    lineno++;
+    if ((lineno==1)&&(sscanf(buffer,"%lf %lf",&dtmp,&dtmp2)==1)){
+      /* First line is single number. Perhaps we have a Siesta EIG file? */
+      if (!fgets(buffer,LINE_SIZE,infile)) error_exit("Unexpected EOF");
+      if (sscanf(buffer,"%d %d %d %lf",&no,&nspins,&nk,&dtmp2)==3){
+	e->e_fermi=malloc(sizeof(double));
+	if (!e->e_fermi) error_exit("malloc error");
+	*(e->e_fermi)=dtmp;
+	siesta=1;
+	break;
+      }
+    }
     ptr=strstr(buffer,"Fermi");
     if ((ptr)&&((ptr-buffer)<6)){
       scale=0;
-      ptr2=index(buffer,'=');
+      ptr2=strchr(buffer,'=');
       ptr=strstr(buffer,"hartree");
       if ((ptr)&&(ptr<ptr2)) scale=H_eV;
       ptr=strstr(buffer,"eV");
@@ -896,7 +956,7 @@ void abinit_eig_read(FILE* infile, struct unit_cell *c, struct contents *m,
     ptr=strstr(buffer,"Eigenvalues");
     if ((ptr)&&((ptr-buffer)<6)){
       scale=0;
-      ptr2=index(buffer,'=');
+      ptr2=strchr(buffer,'=');
       ptr=strstr(buffer,"hartree");
       if ((ptr)&&(ptr<ptr2)) scale=H_eV;
       ptr=strstr(buffer,"eV");
@@ -910,7 +970,7 @@ void abinit_eig_read(FILE* infile, struct unit_cell *c, struct contents *m,
       if (sscanf(ptr2+1,"%d",&itmp)==1){
 	k->n=itmp;
 	k->kpts=malloc(itmp*sizeof(struct atom));
-	if (!k->kpts) error_exit("Malloc error for kpoints");
+	if (!k->kpts) error_exit("malloc error for kpoints");
 	}
       else error_exit("Failed to read number of kpts");
 
@@ -923,6 +983,54 @@ void abinit_eig_read(FILE* infile, struct unit_cell *c, struct contents *m,
     }
   }
 
+  if (siesta){
+    fprintf(stderr,"Siesta EIG file detected\n");
+    if (nspins!=1) error_exit("cannot read Siesta EIG file for nspins!=1");
+    e->nbands=no;
+    if (k->n==0){
+      /* Need to read a kpoints file */
+      newfile=strrsubs(dict_get(m->dict,"in_file"),"EIG","KP");
+      f=fopen(newfile,"r");
+      if (f){
+	fprintf(stderr,"Additionally reading %s\n",newfile);
+	siesta_kp_read(f,k);
+	fclose(f);
+      }
+      else{
+	fprintf(stderr,"Need kpoints, but failed to find .KP file\n");
+	exit(1);
+      }
+    }
+    if (k->n!=nk){
+      fprintf(stderr,"Confused: are there %d or %d kpoints?\n",k->n,nk);
+      exit(1);
+    }
+    e->eval=malloc(nk*no*sizeof(double));
+    if (!e->eval) error_exit("malloc error for eigenvalues");
+    for(i=0;i<nk;i++){
+      if (!fgets(buffer,SLINE_SIZE,infile)) error_exit("Unexpected EOF");
+      lineno++;
+      if (sscanf(buffer,"%d%n",&j,&itmp)!=1) error_exit("Parse error");
+      if (j!=i+1) {
+	fprintf(stderr,"Aborting, confused: out of order kpoints "
+		"found %d expected %d line %d\n",
+		j,i+1,lineno);
+	fprintf(stderr,"%s\n",buffer);
+	exit(1);
+      }
+      ptr=buffer+itmp;
+      for(j=0;j<no;j++){
+	while (sscanf(ptr,"%lf%n",e->eval+i*e->nbands+j,&itmp)!=1){
+	  if (!fgets(buffer,SLINE_SIZE,infile)) error_exit("Unexpected EOF");
+	  lineno++;
+	  ptr=buffer;
+	}
+	ptr+=itmp;
+      }
+    }
+    return;
+  }
+  
   if (debug) fprintf(stderr,"Reading %d kpoints with %d spins\n",k->n,
 		     e->nspins);
 
@@ -940,7 +1048,7 @@ void abinit_eig_read(FILE* infile, struct unit_cell *c, struct contents *m,
 	e->nbands=dummy;
 	if (debug) fprintf(stderr,"Reading %d bands\n",e->nbands);
 	e->eval=malloc(e->nbands*e->nspins*k->n*sizeof(double));
-	if (!e->eval) error_exit("Malloc error for eigenvalues");
+	if (!e->eval) error_exit("malloc error for eigenvalues");
       }
       if (dummy!=e->nbands) error_exit("Varying number of bands");
       if (isp==0){
