@@ -1,13 +1,13 @@
 /* Read data from an Abinit 8.x Fortran binary file.
  *
  * Unclear which versions of abinit this might work with
- * tested with 8.6.3, 8.10.x, 9.0.2
+ * tested with 8.6.3, 8.10.x, 9.0.2, 9.6.2
  *
  * Defaults to rescaling things recognised as densities from Bohr to A.
  * Does not if "raw" specified (-R flag)
  */
 
-/* Copyright (c) 2018 MJ Rutter 
+/* Copyright (c) 2018--2023 MJ Rutter 
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -31,10 +31,11 @@
  *
  * the occupation of a band is 2 if doubly occupied
  *
- * Note https://docs.abinit.org/guide/abinit/#64-the-header is
- * (currently) wrong. The format is
+ * See https://docs.abinit.org/guide/abinit/#54-the-header but beware
+ * of errors. Abinit 8.6 writes something closer to what that page describes
+ * as the version 9.x format. The format is
  *
- * Record 1: as URL
+ * Record 1: as URL (codvsn length varies)
  * Record 2: Ditto
  * Record 3: Ditto
  * Record 4: Atomic positions and total energy
@@ -56,7 +57,7 @@ void abinit_header_read(FILE* infile, struct unit_cell *c,
 		        struct es *elect,
 		        int fft[3], int **gamma, int *fileform){
   int i,j,tmp,reclen;
-  char version[9];
+  char version[9],junk[12];
   int headform;
   int natoms,nsym,npsp,usepaw,ntypes,bantot,*nband,nbands;
   long offset;
@@ -74,6 +75,7 @@ void abinit_header_read(FILE* infile, struct unit_cell *c,
   if ((tmp!=14)&&(tmp!=16))
     error_exit("Unexpected first record length in abinit_read");
   fread(version,tmp-8,1,infile);
+  version[tmp-8]=0;
   fread(&headform,4,1,infile);
   fread(fileform,4,1,infile);
   tmp=0;
@@ -82,7 +84,8 @@ void abinit_header_read(FILE* infile, struct unit_cell *c,
     error_exit("Unexpected first record length in abinit_read");
   version[8]=0;
   if (debug) fprintf(stderr,"Abinit version %6s\n",version);
-
+  if (debug>1) fprintf(stderr,"fform=%d\n",*fileform);
+  
   if (headform<80)
     error_exit("Unable to read files from abinit <8.0");
   
@@ -91,7 +94,8 @@ void abinit_header_read(FILE* infile, struct unit_cell *c,
   fread(&reclen,4,1,infile);
   offset=ftell(infile);
   fread(&bantot,4,1,infile);
-  fseek(infile,12,SEEK_CUR);
+  //  fseek(infile,12,SEEK_CUR);
+  fread(junk,12,1,infile);
   fread(&natoms,4,1,infile);
   fread(fft,12,1,infile);
   fread(&kp->n,4,1,infile);
@@ -101,7 +105,8 @@ void abinit_header_read(FILE* infile, struct unit_cell *c,
   fread(&nsym,4,1,infile);
   fread(&npsp,4,1,infile);
   fread(&ntypes,4,1,infile);
-  fseek(infile,2*4,SEEK_CUR);
+  //  fseek(infile,2*4,SEEK_CUR);
+  fread(junk,8,1,infile);
   fread(&usepaw,4,1,infile);
   fread(&elect->cut_off,8,1,infile);
   elect->cut_off*=H_eV;
@@ -172,7 +177,8 @@ void abinit_header_read(FILE* infile, struct unit_cell *c,
 
   fread(&reclen,4,1,infile);
   offset=ftell(infile);
-  fseek(infile,8,SEEK_CUR);
+  //  fseek(infile,8,SEEK_CUR);
+  fread(junk,8,1,infile);
   posn=malloc(natoms*3*sizeof(double));
   if (!posn) error_exit("Malloc error for atomic positions");
   fread(posn,natoms*3*sizeof(double),1,infile);
@@ -253,148 +259,167 @@ void abinit_header_read(FILE* infile, struct unit_cell *c,
 void abinit_charge_read(FILE* infile, struct unit_cell *c,
 			struct contents *m, struct kpts *kp,
                         struct grid *gptr, struct es *elect){
-  int i,j,k;
-  int reclen,fft[3],fileform,*gamma;
+  int i,j,k,ns,dummy;
+  int reclen,fft[3],fft_size,fileform,*gamma;
   double scale,*ptr,*chgden;
 
+  gamma=NULL;
   abinit_header_read(infile,c,m,kp,elect,fft,&gamma,&fileform);
-
+  if (gamma) free(gamma);
+  
   if (((fileform>=50)&&(fileform<100))&&
       ((!(flags&CHDEN))&&(!(flags&SPINDEN)))) return;
+
+  fft_size=fft[0]*fft[1]*fft[2];
+  chgden=NULL;
   
   fread(&reclen,4,1,infile);
-  if (reclen!=8*fft[0]*fft[1]*fft[2]){
+  if (reclen!=8*fft_size){
     fprintf(stderr,"Volumetric data not understood, so not reading\n");
     fprintf(stderr,"Expected %d bytes, found %d\n",
-            8*fft[0]*fft[1]*fft[2],reclen);
+            8*fft_size,reclen);
     return;
   }
 
-  gptr=grid_new(gptr);
-
-  for(i=0;i<3;i++) gptr->size[i]=fft[i];
-  if(!(gptr->data=malloc(gptr->size[0]*gptr->size[1]*
-                         gptr->size[2]*sizeof(double))))
-    error_exit("Malloc error for grid data");
-
-  
-  if (fread(gptr->data,1,reclen,infile)!=reclen){
-    fprintf(stderr,"Error reading grid data: ignoring grid\n");
-    free(gptr->data);
-    gptr->data=NULL;
-    return;
+  if ((fileform==52)&&(flags&SPINDEN)&&(elect->nspins==4)&&
+      (!(flags&CHDEN))){ /* skip charge */
+    fseek(infile,reclen+4,SEEK_CUR);
   }
+  else{
+    gptr=grid_new(gptr);
+    for(i=0;i<3;i++) gptr->size[i]=fft[i];
+    if(!(ptr=malloc(fft_size*sizeof(double))))
+      error_exit("Malloc error for grid data");
 
-  fseek(infile,4,SEEK_CUR);
-  
-  /* Transpose data grid from Fortran to C ordering */
-
-  if(!(ptr=malloc(gptr->size[0]*gptr->size[1]*
-                         gptr->size[2]*sizeof(double))))
-    error_exit("Malloc error for temporary data");
-
-  for(i=0;i<fft[0];i++)
-    for(j=0;j<fft[1];j++)
-      for(k=0;k<fft[2];k++)
-        ptr[k+j*fft[2]+i*fft[2]*fft[1]]=gptr->data[k*fft[0]*fft[1]+j*fft[0]+i];
-  free(gptr->data);
-  gptr->data=ptr;
-  
-  if ((fileform>=50)&&(fileform<100)){
-    gptr->name="Density";
-    if (!(flags&RAW)){
-      if (debug) fprintf(stderr,"Rescaling data from Bohr^-3 to A^-3\n");
-      scale=1/(BOHR*BOHR*BOHR);
-      for(i=0;i<gptr->size[0]*gptr->size[1]*gptr->size[2];i++)
-        gptr->data[i]*=scale;
+    if (fread(ptr,1,reclen,infile)!=reclen){
+      fprintf(stderr,"Error reading grid data: ignoring grid\n");
+      free(ptr);
+      return;
     }
-  }
-  else if ((fileform>=100)&&(fileform<120)) {
-    gptr->name="Potential";
-    if (fileform==104)
-      gptr->name="Potential_VHA";
-    else if (fileform==105)
-      gptr->name="Potential_PSP";
-    else if (fileform==106)
-      gptr->name="Potential_VCLMB";
-    else if (fileform==107)
-      gptr->name="Potential_VHXC";
-    else if (fileform==107)
-      gptr->name="Potential_VXC";
+
+    /*    fseek(infile,4,SEEK_CUR);  */
+    fread(&dummy,4,1,infile);
+  
+    /* Transpose data grid from Fortran to C ordering */
+
+    if(!(gptr->data=malloc(fft_size*sizeof(double))))
+      error_exit("Malloc error for temporary data");
+
+    for(i=0;i<fft[0];i++)
+      for(j=0;j<fft[1];j++)
+	for(k=0;k<fft[2];k++)
+	  gptr->data[k+j*fft[2]+i*fft[2]*fft[1]]=
+	    ptr[k*fft[0]*fft[1]+j*fft[0]+i];
+    free(ptr);
+  
+    if ((fileform>=50)&&(fileform<100)){
+      gptr->name="Density";
+      if (!(flags&RAW)){
+	if (debug) fprintf(stderr,"Rescaling data from Bohr^-3 to A^-3\n");
+	scale=1/(BOHR*BOHR*BOHR);
+	for(i=0;i<fft_size;i++)
+	  gptr->data[i]*=scale;
+      }
+    }
+    else if ((fileform>=100)&&(fileform<120)) {
+      gptr->name="Potential";
+      if (fileform==104)
+	gptr->name="Potential_VHA";
+      else if (fileform==105)
+	gptr->name="Potential_PSP";
+      else if (fileform==106)
+	gptr->name="Potential_VCLMB";
+      else if (fileform==107)
+	gptr->name="Potential_VHXC";
+      else if (fileform==107)
+	gptr->name="Potential_VXC";
       
-    if (!(flags&RAW)){
-      if (debug) fprintf(stderr,"Rescaling data from Ha to eV\n");
-      scale=H_eV;
-      for(i=0;i<gptr->size[0]*gptr->size[1]*gptr->size[2];i++)
-        gptr->data[i]*=scale;
+      if (!(flags&RAW)){
+	if (debug) fprintf(stderr,"Rescaling data from Ha to eV\n");
+	scale=H_eV;
+	for(i=0;i<fft_size;i++)
+	  gptr->data[i]*=scale;
+      }
     }
+    else gptr->name="Unknown";
   }
-  else gptr->name="Unknown";
-
-  if ((fileform!=52)||(elect->nspins!=2)) return;
+  
+  if ((fileform!=52)||(elect->nspins==1)) return;
   if (!(flags&SPINDEN)) return;
 
   /* We have some spin to read */
 
   fread(&reclen,4,1,infile);
-  if (reclen!=8*fft[0]*fft[1]*fft[2]){
+  if (reclen!=8*fft_size){
     fprintf(stderr,"Spin data not understood, so not reading\n");
     fprintf(stderr,"Expected %d bytes, found %d\n",
-            8*fft[0]*fft[1]*fft[2],reclen);
+            8*fft_size,reclen);
     return;
   }
   if (!(ptr=malloc(reclen)))
     error_exit("Malloc error for temporary data");
-  if (fread(ptr,1,reclen,infile)!=reclen){
-    fprintf(stderr,"Error reading temporary spin data: ignoring\n");
-    free(ptr);
-    return;
-  }
 
-  chgden=gptr->data;
-  if (flags&CHDEN){
+  if (elect->nspins==2){
+    chgden=gptr->data;
+    if (flags&CHDEN) gptr=grid_new(gptr);
+  }
+  else
     gptr=grid_new(gptr);
-    for(i=0;i<3;i++) gptr->size[i]=fft[i];
-    if(!(gptr->data=malloc(gptr->size[0]*gptr->size[1]*
-                           gptr->size[2]*sizeof(double))))
+  
+  for(i=0;i<3;i++) gptr->size[i]=fft[i];
+  if (!(gptr->data=malloc((elect->nspins-1)*fft_size*sizeof(double))))
       error_exit("Malloc error for spin data");
-  }
-  else{ /* Discard charge density data that was not requested 
-         * by overwriting its grid entry */
-    if(!(gptr->data=malloc(gptr->size[0]*gptr->size[1]*
-                           gptr->size[2]*sizeof(double))))
-      error_exit("Malloc error for spin data");
-  }    
 
+  for(ns=0;ns<elect->nspins-1;ns++){
+    if (ns) fread(&dummy,4,1,infile);
+    if (fread(ptr,1,reclen,infile)!=reclen)
+      error_exit("Error reading spin data");
+  
   /* Transpose data grid from Fortran to C ordering */
 
-  for(i=0;i<fft[0];i++)
-    for(j=0;j<fft[1];j++)
-      for(k=0;k<fft[2];k++)
-        gptr->data[k+j*fft[2]+i*fft[2]*fft[1]]=ptr[k*fft[0]*fft[1]+j*fft[0]+i];
+    for(i=0;i<fft[0];i++)
+      for(j=0;j<fft[1];j++)
+	for(k=0;k<fft[2];k++)
+	  gptr->data[k+j*fft[2]+i*fft[2]*fft[1]+ns*fft_size]=
+	    ptr[k*fft[0]*fft[1]+j*fft[0]+i];
+
+    fread(&dummy,4,1,infile);
+
+  }
   free(ptr);
   
-  if (!(flags&RAW)){
-    if (debug) fprintf(stderr,"Rescaling spin from Bohr^-3 to A^-3 "
-                       "and calculating net spin\n");
-    scale=2/(BOHR*BOHR*BOHR);
-    for(i=0;i<gptr->size[0]*gptr->size[1]*gptr->size[2];i++)
-      gptr->data[i]=gptr->data[i]*scale-chgden[i];
+  if (elect->nspins==2){
+    gptr->name="Spin";
+    if (!(flags&RAW)){
+      if (debug) fprintf(stderr,"Rescaling spin from Bohr^-3 to A^-3\n"
+			 "and calculating net spin\n");
+      scale=2/(BOHR*BOHR*BOHR);
+      for(i=0;i<fft_size;i++)
+	gptr->data[i]=gptr->data[i]*scale-chgden[i];
+      if (!(flags&CHDEN)) free(chgden);
+    }
+  }
+  else if (elect->nspins==4){
+    gptr->name="Vector spin";
+    gptr->comps=3;
+    if (!(flags&RAW)){
+    if (debug) fprintf(stderr,"Rescaling spin from Bohr^-3 to A^-3\n");
+    scale=1/(BOHR*BOHR*BOHR);
+    for(i=0;i<3*fft_size;i++)
+      gptr->data[i]*=scale;
+    }
   }
-  if (!(flags&CHDEN)) free(chgden);
-
-  gptr->name="Spin";
-
 }
 
 
 void abinit_psi_read(FILE* infile, struct unit_cell *c,
                      struct contents *m, struct kpts *kp, struct grid *gptr,
                      struct es *elect, int *i_grid){
-  int i,j,ikpt,nb,isppol;
+  int i,j,ikpt,nb,isppol,junk;
   int reclen,fft[3],fileform,*gamma;
   int npw,nspinor,ispinor,nbnd,off;
   int *pwgrid;
+  long bytes_to_skip;
   double *dptr;
 
   dict_add(m->dict,"band_read_order",NULL); /* Delete any old entry */
@@ -435,14 +460,14 @@ void abinit_psi_read(FILE* infile, struct unit_cell *c,
         fprintf(stderr,"Gamma point storage for this kpt\n");
       fread(&nspinor,4,1,infile);
       fread(&nbnd,4,1,infile);
-      fseek(infile,4,SEEK_CUR);
+      fread(&junk,1,4,infile);
       if (nbnd>elect->nbands) error_exit("nbnd for kpt > nbands");
       pwgrid=realloc(pwgrid,3*npw*sizeof(int));
       if (!pwgrid) error_exit("Malloc error for pwgrid");
       fread(&reclen,4,1,infile);
       if (reclen!=12*npw) error_exit("Unexpected record length for npw");
       fread(pwgrid,4,3*npw,infile);
-      fseek(infile,4,SEEK_CUR);
+      fread(&junk,1,4,infile);
       for(i=0;i<3;i++)
         for(j=0;j<npw;j++)
           if (pwgrid[3*j+i]<0) pwgrid[3*j+i]+=fft[i];
@@ -454,18 +479,23 @@ void abinit_psi_read(FILE* infile, struct unit_cell *c,
       /* Abinit has band occupancy of two if nbspins=1 */
       if (elect->nbspins==1)
         for(i=0;i<nbnd;i++) elect->occ[off+i]*=0.5;
-      fseek(infile,4,SEEK_CUR);
+      fread(&junk,1,4,infile);
+      bytes_to_skip=0;
       for(nb=0;nb<nbnd;nb++){
-        i=fread(&reclen,4,1,infile);
-        if (reclen!=16*npw*nspinor){
-          error_exit("Unexpected record length for band");
-        }
         if ((flags&BANDPARITY||((flags&BANDS)&&
 				((inrange(nb+1,elect->band_range))&&
 				 (inrange(ikpt+1,elect->kpt_range))&&
 				 ((elect->nbspins==1)||
                                   (inrange(isppol,elect->spin_range))))))){
           if (debug>1) fprintf(stderr,"Reading band %d\n",nb+1);
+	  if (bytes_to_skip){
+	    fseek(infile,bytes_to_skip,SEEK_CUR);
+	    bytes_to_skip=0;
+	  }
+	  i=fread(&reclen,4,1,infile);
+	  if (reclen!=16*npw*nspinor){
+	    error_exit("Unexpected record length for band");
+	  }
           for(ispinor=1;ispinor<=nspinor;ispinor++){
             dptr=malloc(16*npw);
             if (!dptr) error_exit("Malloc error for band");
@@ -485,8 +515,12 @@ void abinit_psi_read(FILE* infile, struct unit_cell *c,
 	   free(dptr); 
           } /* end spinor loop */
         }
-        else fseek(infile,reclen+4,SEEK_CUR);
+        else bytes_to_skip+=16*npw*nspinor+8;
       } /* end for bands */
+      if (bytes_to_skip){
+	fseek(infile,bytes_to_skip,SEEK_CUR);
+	bytes_to_skip=0;
+      }
     }
   }
   if (pwgrid) free(pwgrid);
