@@ -24,13 +24,6 @@
 
 #include "c2xsf.h"
 
-/* cart2abc requires this funct from super.c */
-void grid_interp(struct unit_cell *c, struct grid *grid,int old_fft[3],
-   double old_basis[3][3],double old_recip[3][3]);
-
-/* From primitive.c */
-void shorten(double vfinal[3][3]);
-
 /* Check whether set is right- or left-handed */
 int is_rhs(double b[3][3]){
   if ((b[0][0]*(b[1][1]*b[2][2]-b[1][2]*b[2][1])
@@ -206,8 +199,34 @@ void abc2cart(double *abc, struct unit_cell *c){
 
 }
 
+void vabs2frac(double v[3],double recip[3][3]){
+  int i,j;
+  double vtmp[3];
+
+  for(i=0;i<3;i++){
+    vtmp[i]=0;
+    for(j=0;j<3;j++)
+      vtmp[i]+=v[j]*recip[i][j];
+  }
+  for(i=0;i<3;i++)
+    v[i]=vtmp[i];
+}
+
+void vfrac2abs(double v[3],double basis[3][3]){
+  int i,j;
+  double vtmp[3];
+
+  for(i=0;i<3;i++){
+    vtmp[i]=0;
+    for(j=0;j<3;j++)
+      vtmp[i]+=v[j]*basis[j][i];
+  }
+  for(i=0;i<3;i++)
+    v[i]=vtmp[i];
+}
+
 /* Convert from cartesian basis set to a,b,c,alpha,beta,gamma */
-/* If m and gptr are NULL, and fix==0, just calculate a,b,c,alpha,beta,gamma */
+/* If m and gptr are NULL, just calculate a,b,c,alpha,beta,gamma */
 /* If m not NULL, rotate axes to a along x, b in x-y plane, etc */
 
 /* Note: basis2abc implies a rotation in Cartesian space if one
@@ -225,8 +244,9 @@ void abc2cart(double *abc, struct unit_cell *c){
  */
 
 void cart2abc(struct unit_cell *c, struct contents *m, double *abc, 
-              struct grid *gptr, int fix){
-
+              struct grid *gptr){
+  int i;
+  
   if (debug>2) fprintf(stderr,"cart2abc called with motif %s\n",
                        m?"present":"absent");
 
@@ -237,18 +257,34 @@ void cart2abc(struct unit_cell *c, struct contents *m, double *abc,
 
   if (m){
     if (is_rhs(c->basis)==0) make_rhs(c,m,abc,gptr);
+    /* Correct Cartesian vector atomic quantities */
+    for(i=0;i<m->n;i++){
+      vabs2frac(m->atoms[i].force,c->recip);
+      vabs2frac(m->atoms[i].v,c->recip);
+      vabs2frac(m->atoms[i].vspin,c->recip);
+    }
     abc2cart(abc,c);
     real2rec(c);
     addabs(m->atoms,m->n,c->basis);
+    for(i=0;i<m->n;i++){
+      vfrac2abs(m->atoms[i].force,c->basis);
+      vfrac2abs(m->atoms[i].v,c->basis);
+      vfrac2abs(m->atoms[i].vspin,c->basis);
+    }
   }
 
+  if (c->primitive){
+    free_cell(c->primitive);
+    c->primitive=NULL;
+  }
+  
 }
 
 /* As cart2abc, but also correct symmetry operations for any rotation
    which occurs */
 
 void cart2abc_sym(struct unit_cell *c, struct contents *m, double *abc, 
-		  struct grid *gptr, int fix, struct symmetry *s){
+		  struct grid *gptr, struct symmetry *s){
   int i,j,k;
   double basis_t[3][3],recip_t[3][3];
   double (*rot)[3][3],(*tr)[3];
@@ -270,7 +306,7 @@ void cart2abc_sym(struct unit_cell *c, struct contents *m, double *abc,
     }
   }
 
-  cart2abc(c,m,abc,gptr,fix);
+  cart2abc(c,m,abc,gptr);
 
   if (s){
 
@@ -318,7 +354,7 @@ void basis2abc(double b[3][3], double abc[6]){
 /* Make a set of axes into a rhs */
 void make_rhs(struct unit_cell *c, struct contents *m, double *abc,
               struct grid *gptr){
-  int i,j,k,itmp,preserve_c;
+  int i,j,k,ic,itmp,preserve_c,npts;
   double dtmp,(*b)[3],*dptr,*dptr2,*new_grid;
 
   preserve_c=(flags&PRESERVE_C)>>PC_SHIFT;
@@ -376,34 +412,36 @@ void make_rhs(struct unit_cell *c, struct contents *m, double *abc,
       gptr->size[(1+preserve_c)%3]=gptr->size[(2+preserve_c)%3];
       gptr->size[(2+preserve_c)%3]=itmp;
 
-      new_grid=malloc(gptr->size[0]*gptr->size[1]*gptr->size[2]*
-		      sizeof(double));
+      npts=gptr->size[0]*gptr->size[1]*gptr->size[2];
+      new_grid=malloc(npts*gptr->comps*sizeof(double));
       if (!new_grid) error_exit("Malloc error in cart2abc");
       dptr=new_grid;
-      if (preserve_c==2){
-	for(k=0;k<gptr->size[0];k++){
-	  for(j=0;j<gptr->size[1];j++){
-	    dptr2=gptr->data+((j*gptr->size[0])+k)*gptr->size[2];
-	    dptr=new_grid+((k*gptr->size[1])+j)*gptr->size[2];
-	    for(i=0;i<gptr->size[2];i++){
-	      *(dptr++)=*(dptr2++);
+      for(ic=0;ic<gptr->comps;ic++){
+	if (preserve_c==2){
+	  for(k=0;k<gptr->size[0];k++){
+	    for(j=0;j<gptr->size[1];j++){
+	      dptr2=gptr->data+((j*gptr->size[0])+k)*gptr->size[2]+ic*npts;
+	      dptr=new_grid+((k*gptr->size[1])+j)*gptr->size[2]+ic*npts;
+	      for(i=0;i<gptr->size[2];i++){
+		*(dptr++)=*(dptr2++);
+	      }
 	    }
 	  }
 	}
-      }
-      else if (preserve_c==0){
-	for(k=0;k<gptr->size[0];k++){
-	  for(j=0;j<gptr->size[1];j++){
-	    dptr2=gptr->data+k*gptr->size[1]*gptr->size[2];
-	    dptr=new_grid+((k*gptr->size[1])+j)*gptr->size[2];
-	    for(i=0;i<gptr->size[2];i++){
-	      *(dptr++)=*(dptr2+j+i*gptr->size[1]);
+	else if (preserve_c==0){
+	  for(k=0;k<gptr->size[0];k++){
+	    for(j=0;j<gptr->size[1];j++){
+	      dptr2=gptr->data+k*gptr->size[1]*gptr->size[2]+ic*npts;
+	      dptr=new_grid+((k*gptr->size[1])+j)*gptr->size[2]+ic*npts;
+	      for(i=0;i<gptr->size[2];i++){
+		*(dptr++)=*(dptr2+j+i*gptr->size[1]);
+	      }
 	    }
 	  }
 	}
+	else error_exit("Unsupported value of preserve_c in make_rhs");
       }
-      else error_exit("Unsupported value of preserve_c in grid_interp");
-
+    
       free(gptr->data);
       gptr->data=new_grid;
 	  
@@ -456,6 +494,7 @@ void init_atoms(struct atom *a, int n){
     for(j=0;j<3;j++) a[i].frac[j]=0;
     for(j=0;j<3;j++) a[i].force[j]=0;
     for(j=0;j<3;j++) a[i].v[j]=0;
+    for(j=0;j<3;j++) a[i].vspin[j]=0;
     a[i].wt=0;
     a[i].spin=0;
     a[i].chg=0;
@@ -490,6 +529,10 @@ void vacuum_adjust(struct unit_cell *c, struct contents *m, double new_abc[3]){
   real2rec(c);
   c->vol=fabs(c->vol);
   addfrac(m->atoms,m->n,c->recip);
+  if (c->primitive){
+    free_cell(c->primitive);
+    c->primitive=NULL;
+  }
 
 }
 
@@ -497,6 +540,10 @@ void vacuum_adjust(struct unit_cell *c, struct contents *m, double new_abc[3]){
 void old_in_new(double old_basis[3][3],double new_recip[3][3]){
   int i,j,k;
   double dot;
+  char *fmt;
+
+  fmt="%g";
+  if (flags&HIPREC) fmt="%.14g";
   
   for(i=0;i<3;i++){
     fprintf(stderr,"(");
@@ -504,10 +551,8 @@ void old_in_new(double old_basis[3][3],double new_recip[3][3]){
       dot=0;
       for(k=0;k<3;k++)
         dot+=old_basis[i][k]*new_recip[j][k];
-      if (aeq(dot,floor(dot+0.5)))
-        fprintf(stderr,"%d",(int)floor(dot+0.5));
-      else
-        fprintf(stderr,"*");
+
+      fprintf(stderr,fmt,dot);
       if (j!=2)fprintf(stderr,",");
     }
     fprintf(stderr,")");
@@ -516,12 +561,41 @@ void old_in_new(double old_basis[3][3],double new_recip[3][3]){
 
 }
 
+void print_old_in_new(double old_basis[3][3], double new_basis[3][3]){
+  int i,j;
+  double buff[9];
+  struct unit_cell c;
+
+  init_cell(&c);
+  c.basis=(void*)buff;
+  
+  for(i=0;i<3;i++)
+    for(j=0;j<3;j++)
+      c.basis[i][j]=new_basis[i][j];
+
+  real2rec(&c);
+
+  fprintf(stderr,"Old basis in terms of new: ");
+  old_in_new(old_basis,c.recip);
+
+  for(i=0;i<3;i++)
+    for(j=0;j<3;j++)
+      c.basis[i][j]=old_basis[i][j];
+  
+  real2rec(&c);
+
+  fprintf(stderr,"New basis in terms of old: ");
+  old_in_new(new_basis,c.recip);
+}
+
 void cell_check(struct unit_cell *c, struct contents *m){
   int i,j,k,ii,n1,n2;
   double min_dist,dtmp,vec[3],frac[3];
   struct unit_cell compact_cell;
 
   n1=n2=0;
+
+  if (!c->basis) return;
   
   if (fabs(c->vol)<2)
     fprintf(stderr,"*** WARNING: surprisingly small cell volume %g\n",
@@ -623,6 +697,43 @@ void addspec(struct contents *m){
   
 }
 
+void add_primitive(struct unit_cell *c, struct contents *m){
+#ifdef SPGLIB
+  int i,j;
+  double old_basis[3][3],old_recip[3][3];
+#endif
+  c->primitive=malloc(sizeof(struct unit_cell));
+  if (!c->primitive) error_exit("malloc error in add_primitive for cell");
+  init_cell(c->primitive);
+  c->primitive->basis=malloc(9*sizeof(double));
+  if (!c->primitive->basis)
+    error_exit("malloc error in add_primitive for vectors");
+  
+#ifdef SPGLIB
+  for(i=0;i<3;i++){
+    for(j=0;j<3;j++){
+      old_basis[i][j]=c->basis[i][j];
+      old_recip[i][j]=c->recip[i][j];
+    }
+  }
+
+  cspg_op(c,m,NULL,NULL,CSPG_PRIM_LATT,tol);
+
+  for(i=0;i<3;i++){
+    for(j=0;j<3;j++){
+      c->primitive->basis[i][j]=c->basis[i][j];
+      c->basis[i][j]=old_basis[i][j];
+      c->primitive->recip[i][j]=c->recip[i][j];
+      c->recip[i][j]=old_recip[i][j];
+    }
+  }
+#else
+  primitive(c,m,c->primitive->basis);
+  real2rec(c->primitive);
+#endif
+}  
+  
+  
 struct grid *grid_new(struct grid *gptr){
   if (gptr->next) gptr=gptr->next;
   gptr->next=malloc(sizeof(struct grid));
@@ -636,12 +747,25 @@ void init_cell(struct unit_cell *c){
   
   c->basis=NULL;
   c->stress=NULL;
+  c->primitive=NULL;
+  c->fbz=NULL;
+  c->ibz=NULL;
+  c->ws_cell=NULL;
   c->vol=0;
   for(i=0;i<3;i++)
     for(j=0;j<3;j++)
       c->recip[i][j]=0.0;
 }
 
+void free_cell(struct unit_cell *c){
+
+  if (!c) return;
+  if (c->primitive) free_cell(c->primitive);
+  if (c->stress) free(c->stress);
+  if (c->basis) free(c->basis);
+  free(c);
+}
+
 void init_elect(struct es *e){
   e->band_range="-";
   e->kpt_range="1";
@@ -663,6 +787,9 @@ void init_elect(struct es *e){
   e->eval=NULL;
   e->nel=e->nup_minus_down=0;
   e->max_nplwv=0;
+  e->path_eval=e->path_occ=NULL;
+  e->path_kpt=NULL;
+  e->path_nbands=0;
 }
 
 void init_motif(struct contents *m){
@@ -686,6 +813,7 @@ void init_grid(struct grid *g){
   int i;
   
   for(i=0;i<3;i++) g->size[i]=0;
+  g->comps=1;
   g->next=NULL;
   g->data=NULL;
   g->name=NULL;
@@ -709,6 +837,7 @@ void init_kpts(struct kpts *k){
 void init_sym(struct symmetry *s){
   s->tol=NULL;
   s->ops=NULL;
+  s->gen=NULL;
   s->n=0;
 }
 
