@@ -1,6 +1,6 @@
 /* Interface between c2x and spglib */
 
-/* Copyright (c) 2014 -- 2019 MJ Rutter 
+/* Copyright (c) 2014 -- 2022 MJ Rutter 
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -28,7 +28,6 @@
 
 #include "c2xsf.h"
 
-void vcross(double a[3],double b[3],double c[3]); /* From ident_sym.c */
 int is_rhs(double b[3][3]); /* From basis.c */
 
 void rotate(double v[3], double axis[3], double theta);
@@ -320,7 +319,7 @@ int cspg_op(struct unit_cell *c, struct contents *m, struct symmetry *s,
 		old_basis[i][1],old_basis[i][2]);
     }
 
-    super(c,m,old_basis,NULL,s,NULL,0);
+    super(c,m,old_basis,NULL,s,NULL,0,NULL);
 
     free(c2->basis);
     free(c2);
@@ -404,6 +403,10 @@ int cspg_op(struct unit_cell *c, struct contents *m, struct symmetry *s,
     else{
       m->n=natoms2;
       s->n=0;
+      if (c->primitive){
+	free_cell(c->primitive);
+	c->primitive=NULL;
+      }
     }
   }
 
@@ -417,7 +420,13 @@ int cspg_op(struct unit_cell *c, struct contents *m, struct symmetry *s,
     }
     else{
       m->n=natoms2;
-      s->n=0;
+      if (op&CSPG_STD_IDEAL) {
+	s->n=0;
+	if (c->primitive){
+	  free_cell(c->primitive);
+	  c->primitive=NULL;
+	}
+      }
     }
   }
     
@@ -431,6 +440,10 @@ int cspg_op(struct unit_cell *c, struct contents *m, struct symmetry *s,
     else{
       m->n=natoms2;
       s->n=0;
+      if (c->primitive){
+	free_cell(c->primitive);
+	c->primitive=NULL;
+      }
     }
   }
 
@@ -442,15 +455,22 @@ int cspg_op(struct unit_cell *c, struct contents *m, struct symmetry *s,
       error_exit("spglib finds no primitive cell\n");
     else{
       m->n=natoms2;
-      s->n=0;
     }
   }
       
   if (op&(CSPG_SYM|CSPG_PNT|CSPG_INT|CSPG_SCH)){
-    spg=spg_get_dataset(spg_latt,spg_pos,spg_type,m->n,spg_symprec);
+    if (m->n)
+      spg=spg_get_dataset(spg_latt,spg_pos,spg_type,m->n,spg_symprec);
+    else{
+      vtmp[0]=vtmp[1]=vtmp[2]=0;
+      i=1;
+      spg=spg_get_dataset(spg_latt,&vtmp,&i,1,spg_symprec);
+    }
+    
     if ((tolmin==tol)||(debug))
       spg_sym_print(spg,spg_symprec,op);
-    s->n=spg->n_operations;
+    if (spg) s->n=spg->n_operations;
+    else s->n=0;
   }
   
 
@@ -475,21 +495,34 @@ int cspg_op(struct unit_cell *c, struct contents *m, struct symmetry *s,
     for(i=0;i<3;i++)
       for(j=0;j<3;j++)
 	new_basis[j][i]=spg_latt[i][j];
-    super(c,m,new_basis,kp,NULL,NULL,4);
+    super(c,m,new_basis,kp,NULL,NULL,4,NULL);
   }
-  
+
   for(i=0;i<3;i++)
     for(j=0;j<3;j++)
       c->basis[j][i]=spg_latt[i][j];
   real2rec(c);
 
-  if (debug&&(op&CSPG_PRIM_NR)){
-    fprintf(stderr,"Old basis in terms of new: ");
-    old_in_new(old_basis,c->recip);
+  if (op==CSPG_PRIM_LATT){
+    free(auid);
+    free(spg_type);
+    free(spg_pos);
+    return 0;
+  }
+  
+  if (debug&&(op&CSPG_PRIM_NR))
+    print_old_in_new(old_basis,c->basis);
+
+  /* Should we retain old sym-ops? */
+  
+  if ((op&(CSPG_PRIM_NR|CSPG_STD))&&(s->n)){
+    sym_basis(s,c);
   }
   
   free(m->atoms);
   m->atoms=malloc(m->n*sizeof(struct atom));
+  if (!m->atoms) error_exit("malloc error for atoms in c2x2spg");
+  init_atoms(m->atoms,m->n);
   for(i=0;i<m->n;i++){
     for(j=0;j<3;j++)
       m->atoms[i].frac[j]=spg_pos[i][j];
@@ -510,7 +543,7 @@ int cspg_op(struct unit_cell *c, struct contents *m, struct symmetry *s,
   if ((op&(CSPG_SYM|CSPG_PNT))&&(s->n)) {
     s->ops=sym_frac2abs(spg->rotations,spg->translations,c,s->n);
     if (op&CSPG_LST)
-      for(i=0;i<s->n;i++) ident_sym(&s->ops[i],c,stderr);
+      for(i=0;i<s->n;i++) ident_sym(&s->ops[i],c,m,stderr);
   }
 
   free(auid);
@@ -541,6 +574,9 @@ int cspg_op(struct unit_cell *c, struct contents *m, struct symmetry *s,
 		spg_tr[i][2]);
       }
       /* Translations are "easy" */
+      /* Sometimes SPGlib returns a translation vector component of 1 */
+      for(j=0;j<3;j++)
+	if (aeq(fabs(spg_tr[i][j]),1.0)) spg_tr[i][j]=0;
       if ((spg_tr[i][0]!=0)||(spg_tr[i][1]!=0)||(spg_tr[i][2]!=0)){
 	s[i].tr=malloc(3*sizeof(double));
 	if(!s[i].tr) error_exit("Malloc error for tr in sym_frac2abs");
@@ -593,37 +629,37 @@ int cspg_op(struct unit_cell *c, struct contents *m, struct symmetry *s,
 #else
 int cspg_op(struct unit_cell *c, struct contents *m, struct symmetry *s,
             struct kpts *kp, int op, double tolmin){
-    fprintf(stderr,"Error: spglib functionality not available"
-	    " in this version of c2x\n");
-    exit(1);
-  }
+  fprintf(stderr,"Error: spglib functionality not available"
+	  " in this version of c2x\n");
+  exit(1);
+}
 
-  void cspg_hall2sym(int hall, struct unit_cell *c, struct symmetry *s){
-    if (hall==1) return;
-    fprintf(stderr,"Error: spglib functionality not available"
-	    " so cannot expand spacegroup\n");
-    exit(1);
-  }
+void cspg_hall2sym(int hall, struct unit_cell *c, struct symmetry *s){
+  if (hall==1) return;
+  fprintf(stderr,"Error: spglib functionality not available"
+	  " so cannot expand spacegroup\n");
+  exit(1);
+}
 
 #endif
 
-  /* Does space group have two possible origins? */
-  int spgr_is_double(int spgr){
-    int doubles[]={48,50,59,70,85,86,88,125,126,129,130,133,
-		   134,137,138,141,142,201,203,222,224,227,228,0};
-    int i,hit;
-
-    hit=0;
-    i=0;
-    while(doubles[i]){
-      if (doubles[i]==spgr){
-	hit=1;
-	break;
-      }
-      i++;
+/* Does space group have two possible origins? */
+int spgr_is_double(int spgr){
+  int doubles[]={48,50,59,70,85,86,88,125,126,129,130,133,
+    134,137,138,141,142,201,203,222,224,227,228,0};
+  int i,hit;
+  
+  hit=0;
+  i=0;
+  while(doubles[i]){
+    if (doubles[i]==spgr){
+      hit=1;
+      break;
     }
-    return hit;
+    i++;
   }
+  return hit;
+}
 
 
   void rotate(double v[3], double axis[3], double theta){
@@ -810,3 +846,23 @@ int cspg_op(struct unit_cell *c, struct contents *m, struct symmetry *s,
 
   }
 #endif
+
+
+void niggli_basis(double basis[3][3]){
+
+#ifdef SPGLIB
+  double x[3][3];
+  int i,j;
+  for(i=0;i<3;i++)
+    for(j=0;j<3;j++)
+      x[i][j]=basis[j][i];
+  spg_niggli_reduce(x,tol);
+  for(i=0;i<3;i++)
+    for(j=0;j<3;j++)
+      basis[i][j]=x[j][i];
+#else
+  fprintf(stderr,"Warning: non-niggli basis reduction being applied\n");
+  shorten(basis);
+#endif
+  
+}
