@@ -36,17 +36,25 @@
 static int cellreadline(char *buffer, int len, char *dir,
                        struct infiles **filesp);
 static int cell_read_length(char *buff, double *x);
+static int cell_read_inv_length(char *buff, double *x);
 
 static char *ptr;
 static char **title;
 static int is_onetep;
   
-static void spin_read(char *buff, double *spin){
+static void spin_read(char *buff, struct atom *a){
   while (*buff==' ') buff++;
   if (!strncasecmp(buff,"spin",4)){
     buff+=4;
     while ((*buff==' ')||(*buff=='=')||(*buff==':')) buff++;
-    sscanf(buff,"%lf",spin);
+    if (*buff=='('){
+      buff++;
+      if (sscanf(buff,"%lf %lf %lf",a->vspin,a->vspin+1,a->vspin+2)!=3)
+	fprintf(stderr,"Warning: failed to parse vector spin on atom\n");
+    }
+    else
+      if (sscanf(buff,"%lf",&a->spin)!=1)
+	fprintf(stderr,"Warning: failed to parse spin on atom\n");
   }
 }
 
@@ -83,7 +91,10 @@ void cell_read(FILE* in, struct unit_cell *c, struct contents *m,
   struct label {char *l; char *sym;} *labels;
   struct infiles *files;
   char *dir;
+  char *cell_misc_lst[]={"fix_all_ions","fix_all_cell","fix_vol",
+    "snap_to_symmetry",""};
 
+  
   dir=dict_get(m->dict,"in_dir");
   
   files=malloc(sizeof(struct infiles));
@@ -92,6 +103,8 @@ void cell_read(FILE* in, struct unit_cell *c, struct contents *m,
   files->next=files->last=NULL;
   files->name=NULL;
   files->line=0;
+  files->ret=NULL;
+  files->count=0;
 
   is_onetep=0;
   if (dict_get(m->dict,"cell_is_onetep")) is_onetep=1;
@@ -160,6 +173,7 @@ void cell_read(FILE* in, struct unit_cell *c, struct contents *m,
           exit(1);
         }
       }
+      continue;
     }
 
     if ((!strncasecmp(ptr,"bs_kpoint_mp_grid",17))||
@@ -191,6 +205,7 @@ void cell_read(FILE* in, struct unit_cell *c, struct contents *m,
           exit(1);
         }
       }
+      continue;
     }
 
 
@@ -202,11 +217,13 @@ void cell_read(FILE* in, struct unit_cell *c, struct contents *m,
       /* Increment ptr2 over spaces, = and : */
       while ((*ptr2)&&((*ptr2==' ')||(*ptr2==':')||(*ptr2=='='))) ptr2++;
       dptr=malloc(sizeof(double));
-      if (sscanf(ptr2,"%lf",dptr)!=1){
-        fprintf(stderr,"Warning: error parsing kpoint_mp_spacing\n");
-        free(dptr);
+      if (!dptr) error_exit("Malloc error for single double!");
+      if (!cell_read_inv_length(ptr2,dptr)){
+	fprintf(stderr,"Error parsing:\n%s\n",buffer);
+	free(dptr);
       }
       else kp->spacing=dptr;
+      continue;
     }
     if ((!strncasecmp(ptr,"bs_kpoint_mp_spacing",20))||
         (!strncasecmp(ptr,"bs_kpoints_mp_spacing",21))){
@@ -215,11 +232,13 @@ void cell_read(FILE* in, struct unit_cell *c, struct contents *m,
       /* Increment ptr2 over spaces, = and : */
       while ((*ptr2)&&((*ptr2==' ')||(*ptr2==':')||(*ptr2=='='))) ptr2++;
       dptr=malloc(sizeof(double));
-      if (sscanf(ptr2,"%lf",dptr)!=1){
-        fprintf(stderr,"Warning: error parsing bs_kpoint_mp_spacing\n");
-        free(dptr);
+      if (!dptr) error_exit("Malloc error for single double!");
+      if (!cell_read_inv_length(ptr2,dptr)){
+	fprintf(stderr,"Error parsing:\n%s\n",buffer);
+	free(dptr);
       }
       else kp->bs_spacing=dptr;
+      continue;
     }
     if ((!strncasecmp(ptr,"bs_kpoint_path_spacing",22))||
         (!strncasecmp(ptr,"bs_kpoints_path_spacing",23))){
@@ -228,13 +247,22 @@ void cell_read(FILE* in, struct unit_cell *c, struct contents *m,
       /* Increment ptr2 over spaces, = and : */
       while ((*ptr2)&&((*ptr2==' ')||(*ptr2==':')||(*ptr2=='='))) ptr2++;
       dptr=malloc(sizeof(double));
-      if (sscanf(ptr2,"%lf",dptr)!=1){
-        fprintf(stderr,"Warning: error parsing bs_kpoint_path_spacing\n");
-        free(dptr);
+      if (!dptr) error_exit("Malloc error for single double!");
+      if (!cell_read_inv_length(ptr2,dptr)){
+	fprintf(stderr,"Error parsing:\n%s\n",buffer);
+	free(dptr);
       }
       else kp->path_spacing=dptr;
+      continue;
     }
-      
+    for(i=0;cell_misc_lst[i][0];i++){
+      if (!strncasecmp(ptr,cell_misc_lst[i],strlen(cell_misc_lst[i]))){
+	dict_strcat(m->dict,"cell_misc_list",ptr);
+	dict_strcat(m->dict,"cell_misc_list","\n");
+	continue;
+      }
+    }
+    
 
 /* What remains ought to be a block or a keyword.  */
     if (strncasecmp(ptr,"%block",6)) continue;
@@ -330,7 +358,7 @@ void cell_read(FILE* in, struct unit_cell *c, struct contents *m,
           }
           if(sscanf(buffer,"%*s %lf %lf %lf%n",dptr,
                     dptr+1,dptr+2,&n)>=3)
-            spin_read(buffer+n,&m->atoms[i].spin);
+            spin_read(buffer+n,m->atoms+i);
           else{
             PARSE_ERROR;
             if (debug) fprintf(stderr,"%s\n",ptr); 
@@ -368,7 +396,7 @@ void cell_read(FILE* in, struct unit_cell *c, struct contents *m,
           /* Now need to read co-ordinates */
           if(multi_scan(buffer+n,dptr,3,&n2)==3){
             n+=n2;
-            spin_read(buffer+n,&m->atoms[i].spin);
+            spin_read(buffer+n,m->atoms+i);
           }
           else{
             PARSE_ERROR;
@@ -452,14 +480,14 @@ void cell_read(FILE* in, struct unit_cell *c, struct contents *m,
         cellreadline(buffer,LINE_SIZE,dir,&files);
         if (!strncasecmp(buffer,"%endblock",9)) break;
         kp->bs_kpts=realloc(kp->bs_kpts,(i+1)*sizeof(struct atom));
-        if (!kp->kpts) error_exit("realloc error for kpts");
+        if (!kp->bs_kpts) error_exit("realloc error for kpts");
 	init_atoms(kp->bs_kpts+i,1);
         if (multi_scan(buffer,kp->bs_kpts[i].frac,3,&j)!=3){
           PARSE_ERROR;
           exit(1);
         }
 	/* Weights are optional here */
-        single_scan(buffer+j,&(kp->kpts[i].wt),NULL);
+        single_scan(buffer+j,&(kp->bs_kpts[i].wt),NULL);
         i++;
       }
       kp->bs_n=i;
@@ -488,9 +516,10 @@ void cell_read(FILE* in, struct unit_cell *c, struct contents *m,
         sym_mat=realloc(sym_mat,(nsym+1)*9*sizeof(double));
         sym_disp=realloc(sym_disp,(nsym+1)*3*sizeof(double));
 	if ((!sym_mat)||(!sym_disp)) error_exit("realloc error in cell_read");
+	/* transpose on read */
         for(i=0;i<3;i++){
-          if(sscanf(buffer,"%lf %lf %lf",sym_mat+9*nsym+3*i,
-                   sym_mat+9*nsym+3*i+1,sym_mat+9*nsym+3*i+2)!=3){
+          if(sscanf(buffer,"%lf %lf %lf",sym_mat+9*nsym+i,
+                   sym_mat+9*nsym+i+3,sym_mat+9*nsym+i+6)!=3){
             PARSE_ERROR;
             exit(1);
           }
@@ -583,12 +612,17 @@ void cell_read(FILE* in, struct unit_cell *c, struct contents *m,
     s->ops=malloc(nsym*sizeof(struct sym_op));
     if (!s->ops) error_exit("Malloc error for symops");
     for(i=0;i<nsym;i++){
+      if (vmod2(sym_disp+3*i)<tol*tol) {
+	s->ops[i].tr=NULL;
+      }
+      else{
       /* Translations are relative */
-      s->ops[i].tr=malloc(3*sizeof(double));
-      if (!s->ops[i].tr) error_exit("malloc error for sym tr");
-      for(j=0;j<3;j++) s->ops[i].tr[j]=sym_disp[3*i]*c->basis[0][j]+
-                         sym_disp[3*i+1]*c->basis[1][j]+
-                         sym_disp[3*i+2]*c->basis[2][j];
+	s->ops[i].tr=malloc(3*sizeof(double));
+	if (!s->ops[i].tr) error_exit("malloc error for sym tr");
+	for(j=0;j<3;j++) s->ops[i].tr[j]=sym_disp[3*i]*c->basis[0][j]+
+			   sym_disp[3*i+1]*c->basis[1][j]+
+			   sym_disp[3*i+2]*c->basis[2][j];
+      }
       /* But rotations are absolute */
       for(j=0;j<3;j++)
         for(k=0;k<3;k++)
@@ -605,7 +639,18 @@ void cell_read(FILE* in, struct unit_cell *c, struct contents *m,
     nsym=0;
   }
 
+  /* Fill in MP params if spacing giving */
+
+  if ((kp->spacing)&&(!kp->mp)){
+    kp->mp=malloc(sizeof(struct mp_grid));
+    if (!kp->mp) error_exit("Malloc error for MP grid");
+    basis2abc(c->recip,lat_abc);
+    for(i=0;i<3;i++) kp->mp->grid[i]=1+((lat_abc[i]/(*kp->spacing)));
+    for(i=0;i<3;i++) kp->mp->disp[i]=0;
+  }
+  
   free(labels);
+  free(files);
 
 }
 
@@ -725,3 +770,45 @@ static int cell_read_length(char *buff, double *x){
 }
 
 
+static int cell_read_inv_length(char *buff, double *x){
+  char *p;
+  int n,i;  
+
+  while((*buff)&&((*buff==' ')||(*buff==':')||(*buff=='='))) buff++;
+
+  /*
+   *  n=sscanf(buff,"%lf %ms",x,&p);
+   */
+
+  n=sscanf(buff,"%lf%n",x,&i);
+  n+=sscanfmsn(buff+i,&p,NULL);
+  
+  if (n==0) return 0;
+
+  if (n==1) return 1;
+
+  if (!strncasecmp(p,"1/bohr",6)){
+    (*x)/=BOHR;
+    free(p);
+    return 1;
+  }
+
+  if (!strncasecmp(p,"1/ang",5)){
+    free(p);
+    return 1;
+  }
+
+  if (!strncasecmp(p,"1/nm",4)){
+    (*x)/=10;
+    free(p);
+    return 1;
+  }
+
+  
+  fprintf(stderr,"Unexpected inverse length unit: %s\n",p);
+  free(p);
+  return 0;
+
+}
+
+
