@@ -47,7 +47,7 @@
     if ((!targ)&&!(targ=malloc(len_x))) error_exit("Malloc error in " #x); \
     fread(targ,len_x,1,infile); \
     if (endian && (len_x==4)) reverse4(targ); \
-    fseek(infile,4,SEEK_CUR); \
+    fread(junk,4,1,infile);				    \
     if (debug>2) fprintf(stderr,"Found %s, %d bytes\n", #x, len_x); \
   }
 
@@ -62,7 +62,7 @@
     if ((!targ)&&!(targ=malloc(len_x))) error_exit("Malloc error in " #x); \
     fread(targ,len_x,1,infile); \
     if (endian) reverse8n(targ,len_x/8);		\
-    fseek(infile,4,SEEK_CUR); \
+    fread(junk,4,1,infile);				    \
     if (debug>2) fprintf(stderr,"Found %s, %d bytes\n", #x, len_x); \
   }
 
@@ -85,15 +85,15 @@ static double *cr_cell_vol;
 void check_read(FILE* infile, struct unit_cell *c, struct contents *m,
                 struct kpts *kp, struct symmetry *s, struct grid *gptr,
                 struct es *elect, int *i_grid){
-  int tmp;
+  int tmp,junk[2],okay;
   char head[HEAD_LEN+1];
   char cbuff[CBUFF+1];
-  int i,j,k,density,na,fft[3],ion_sym_len,ps_pot_len,ver_maj,ver_min;
+  int i,j,k,kk,density,na,fft[3],ion_sym_len,ps_pot_len,ver_maj,ver_min;
   int smisc_size;
   double *dptr1,*dptr2,*column,*castep_basis,conv;
   int *nsp,*nsp_max,*nionsp,*nsymops;
   char *ion_sym,*ps_pots;
-  double *ion_pos,*ion_force,*ion_vel,*mag_mom,*sym_mat,*sym_disp;
+  double *ion_pos,*ion_force,*ion_vel,*mag_mom,*sym_mat,*sym_disp,*mag_dir;
   double *species_charge;
   double *kpts_pos,*wkpts,*bs_kpts_pos,*bs_wkpts;
   int *mp_xyz,*bs_mp_xyz;
@@ -111,7 +111,7 @@ void check_read(FILE* infile, struct unit_cell *c, struct contents *m,
   ion_sym_len=8;
   endian=0;
   castep_basis=NULL;
-  ion_pos=ion_force=ion_vel=mag_mom=NULL;
+  ion_pos=ion_force=ion_vel=mag_mom=mag_dir=NULL;
   nsp=nsp_max=nionsp=NULL;
   ps_pots=NULL;
   kpts_pos=bs_kpts_pos=NULL;
@@ -130,6 +130,7 @@ void check_read(FILE* infile, struct unit_cell *c, struct contents *m,
   nspins=1;
   nbands=0;
   c->basis=NULL;
+  conv=1;
 
   if(debug>2) fprintf(stderr,"check_read called with flags=%d\n",flags);
 
@@ -181,7 +182,7 @@ void check_read(FILE* infile, struct unit_cell *c, struct contents *m,
     head[i]=0;
     if (debug>3) fprintf(stderr,"%d: %.4s\n",tmp,head);
     if (i<tmp) fseek(infile,tmp-i,SEEK_CUR);
-    fseek(infile,4,SEEK_CUR); /* capricious */
+    fread(junk,4,1,infile); /* Fortran end of record length marker ignored */
 
     /* Need to find version of file so that we know how long the ionic
      * species symbol records will be. Version is stored in ASCII
@@ -196,7 +197,8 @@ void check_read(FILE* infile, struct unit_cell *c, struct contents *m,
         if (endian) reverse4(&tmp);
         i=(tmp<HEAD_LEN)? tmp : HEAD_LEN;
         fread(head,i,1,infile);
-        fseek(infile,4,SEEK_CUR);
+	if (i<tmp) fseek(infile,tmp-i,SEEK_CUR);
+        fread(junk,4,1,infile);
         head[i]=0;
         j=0;
         ver_maj=0;
@@ -221,6 +223,7 @@ void check_read(FILE* infile, struct unit_cell *c, struct contents *m,
 	SCAN("CELL%IONIC_VELOCITIES",(24*(*nsp)*(*nsp_max)),ion_vel);
 	SCAN("CELL%SPECIES_SYMBOL",(ion_sym_len*(*nsp)),ion_sym);
 	SCAN("CELL%INITIAL_MAGNETIC_MOMENT",(8*(*nsp)*(*nsp_max)),mag_mom);
+	SCAN("CELL%INITIAL_MAGNETISATION_DIR",(24*(*nsp)*(*nsp_max)),mag_dir);
 	SCAN("CELL%IONIC_CHARGE_REAL",(8*(*nsp)),species_charge);
       }
       SCAN8R("CELL%VOLUME",8,cr_cell_vol);
@@ -250,6 +253,7 @@ void check_read(FILE* infile, struct unit_cell *c, struct contents *m,
         if ((kpts_pos)&&(wkpts)){
           if(!(kp->kpts=malloc((*nkpts)*sizeof(struct atom))))
             error_exit("Malloc error for kpts");
+	  init_atoms(kp->kpts,*nkpts);
           for(i=0;i<*nkpts;i++){
             for(j=0;j<3;j++) kp->kpts[i].frac[j]=kpts_pos[3*i+j];
             kp->kpts[i].wt=wkpts[i];
@@ -320,11 +324,13 @@ void check_read(FILE* infile, struct unit_cell *c, struct contents *m,
           if (!strncmp(spin_method,"VECTOR",6)){
             elect->nspinors=2;
 	    elect->nbspins=1;
+# if 0
             if (flags&SPINDEN){
               fprintf(stderr,"Unable to output vector spin density\n"
 		      " resetting flag requesting spin density output\n");
 	      flags^=SPINDEN;
 	    }
+#endif
 	  }
 
         }
@@ -346,7 +352,7 @@ void check_read(FILE* infile, struct unit_cell *c, struct contents *m,
           fread(&cut,8,1,infile);
           if (endian) reverse8(&cut);
           elect->cut_off=cut*H_eV;
-          fseek(infile,4,SEEK_CUR);
+	  fread(junk,4,1,infile);
         }
         else fseek(infile,4+tmp,SEEK_CUR);
         
@@ -511,31 +517,60 @@ void check_read(FILE* infile, struct unit_cell *c, struct contents *m,
                then store as real */
             if (debug>2)
               fprintf(stderr,"We think we have found a charge density grid\n");
-
-            if (flags&CHDEN){
-	      gptr=grid_new(gptr);
-              gptr->name="Density";
-              for(i=0;i<3;i++) gptr->size[i]=fft[i];
-              gptr->data=malloc(8*fft[0]*fft[1]*fft[2]);
-              if (!gptr->data){
-                fprintf(stderr,"Error allocating density grid\n");
-                exit(1);
-              }
-            } /* End if (flags&CHDEN) */
-            if ((nspins==2)&&(flags&SPINDEN)){
-              gptr2=gptr;
-	      gptr2=grid_new(gptr2);
-              gptr2->name="Spin";
-              for(i=0;i<3;i++) gptr2->size[i]=fft[i];
-              gptr2->data=malloc(8*fft[0]*fft[1]*fft[2]);
-              if (!gptr2->data){
-                fprintf(stderr,"Error allocating density grid\n");
-                exit(1);
-              }
-            }
+	    if (flags&REPLACE_RHO){
+	      if ((!gptr)||(!gptr->data))
+		error_exit("Cannot replace rho as no grid data read");
+	      okay=1;
+	      for(i=0;i<3;i++)
+		if (gptr->size[i]!=fft[i]) okay=0;
+	      if (!okay){
+		fprintf(stderr,"Cannot replace rho as grids differ.\n");
+		fprintf(stderr,"Read %dx%dx%d, found %dx%dx%d\n",
+			gptr->size[0],gptr->size[1],gptr->size[2],
+			fft[0],fft[1],fft[2]);
+		exit(1);
+	      }
+	      if (nspins!=1)
+		error_exit("Cannot replace density unless nspins=1");
+	      if (debug) fprintf(stderr,"Found density to replace\n");
+	    }
+	    else{
+	      if (flags&CHDEN){
+		gptr=grid_new(gptr);
+		gptr->name="Density";
+		for(i=0;i<3;i++) gptr->size[i]=fft[i];
+		gptr->data=malloc(sizeof(double)*fft[0]*fft[1]*fft[2]);
+		if (!gptr->data){
+		  fprintf(stderr,"Error allocating density grid\n");
+		  exit(1);
+		}
+	      } /* End if (flags&CHDEN) */
+	      if (((nspins==2)||(nspins==4))&&(flags&SPINDEN)){
+		gptr2=gptr;
+		gptr2=grid_new(gptr2);
+		if (nspins==2) gptr2->name="Spin";
+		else gptr2->name="Vector spin";
+		for(i=0;i<3;i++) gptr2->size[i]=fft[i];
+		gptr2->data=malloc(sizeof(double)*(nspins-1)*
+				   fft[0]*fft[1]*fft[2]);
+		if (!gptr2->data){
+		  fprintf(stderr,"Error allocating density grid\n");
+		  exit(1);
+		}
+		gptr2->comps=nspins-1;
+	      }
+	    }
             if ((flags&CHDEN)||(flags&SPINDEN)){
-              column=malloc(16*fft[2]);
+	      /* Columns are stored as complex, not real */
+	      if ((flags&SPINDEN)&&(nspins==4))
+		column=malloc(48*fft[2]);
+	      else
+		column=malloc(16*fft[2]);
 	      if (!column) error_exit("malloc error for FFT column");
+	      if (flags&REPLACE_RHO){
+		conv=1;
+		if ((!(flags&RAW))&&(cr_cell_vol)) conv=*cr_cell_vol;
+	      }
               while(tmp==(nspins*16*fft[2]+8)){
                 fread(&i,4,1,infile);
                 if (endian) reverse4(&i);
@@ -546,33 +581,53 @@ void check_read(FILE* infile, struct unit_cell *c, struct contents *m,
 		  error_exit("Unexpected FFT column in density\n");
 		}
                 if (flags&CHDEN){
-                  fread(column,16*fft[2],1,infile);
-                  dptr1=gptr->data+((i-1)*fft[1]+(j-1))*fft[2];
-                  dptr2=column;
-                  for(k=0;k<fft[2];k++){*dptr1++=*dptr2;dptr2+=2;}
+		  dptr1=gptr->data+((i-1)*fft[1]+(j-1))*fft[2];
+		  dptr2=column;
+		  if (flags&REPLACE_RHO){
+		    for(k=0;k<fft[2];k++){
+		      *(dptr2++)=*(dptr1++)*conv;
+		      *(dptr2++)=0;
+		    }
+		    if (endian) reverse8n(column,16*fft[2]);
+		    fwrite(column,16*fft[2],1,infile);
+		  }
+		  else{
+		    fread(column,16*fft[2],1,infile);
+		    for(k=0;k<fft[2];k++){*dptr1++=*dptr2;dptr2+=2;}
+		  }
                 }
                 else fseek(infile,16*fft[2],SEEK_CUR);
-                if (nspins==2){
+                if ((nspins==2)||(nspins==4)){
                   if (flags&SPINDEN){
-                    fread(column,16*fft[2],1,infile);
-                    dptr1=gptr2->data+((i-1)*fft[1]+(j-1))*fft[2];
-                    dptr2=column;
-                    for(k=0;k<fft[2];k++){*dptr1++=*dptr2;dptr2+=2;}
+                    fread(column,16*(nspins-1)*fft[2],1,infile);
+		    for(kk=0;kk<nspins-1;kk++){
+		      dptr1=gptr2->data+((i-1)*fft[1]+(j-1))*fft[2]+
+			kk*fft[0]*fft[1]*fft[2];
+		      dptr2=column+2*kk;
+		      for(k=0;k<fft[2];k++){
+			*dptr1++=*dptr2;
+			dptr2+=2*(nspins-1);
+		      }
+		    }
                   }
-                  else fseek(infile,16*fft[2],SEEK_CUR);
+                  else fseek(infile,16*(nspins-1)*fft[2],SEEK_CUR);
                 }
-		else if (nspins==4)
-		  fseek(infile,48*fft[2],SEEK_CUR);
-                fseek(infile,4,SEEK_CUR);
+		else
+		  if (nspins!=1) error_exit("Impossible in check_read");
+                fread(junk,4,1,infile);
                 fread(&tmp,4,1,infile);
                 if (endian) reverse4(&tmp);
               } /* end while(tmp==(spins... ) */
               free(column);
+	      if (flags&REPLACE_RHO){
+		fprintf(stderr,"Charge density replaced\n");
+		exit(0);
+	      }
               fseek(infile,-4,SEEK_CUR);
               /* Correct endianness */
               if (endian){
                 if (flags&CHDEN) reverse8n(gptr->data,fft[0]*fft[1]*fft[2]);
-                if (flags&SPINDEN) reverse8n(gptr2->data,fft[0]*fft[1]*fft[2]);
+                if (flags&SPINDEN) reverse8n(gptr2->data,(nspins-1)*fft[0]*fft[1]*fft[2]);
               }
               /* Scale */
               if (((flags&RAW)==0)&&((gptr)||(gptr2))){
@@ -584,13 +639,14 @@ void check_read(FILE* infile, struct unit_cell *c, struct contents *m,
                 if (flags&CHDEN)
                   for(i=0;i<fft[0]*fft[1]*fft[2];i++) gptr->data[i]*=conv;
                 if (flags&SPINDEN)
-                  for(i=0;i<fft[0]*fft[1]*fft[2];i++) gptr2->data[i]*=conv;
+                  for(i=0;i<(nspins-1)*fft[0]*fft[1]*fft[2];i++)
+		    gptr2->data[i]*=conv;
               }
 	      if ((flags&RAW)&&((gptr)||(gptr2)))
 		add_cmt(m->comment,"Densities unscaled, e per unit cell");
               if (gptr2) gptr=gptr2;
             } /* end if ((flags&CHDEN)||(flags&SPINDEN)) */
-            else fseek(infile,(nspins*16*fft[2]+16)*fft[1]*fft[0]-4,
+            else fseek(infile,(nspins*16L*fft[2]+16)*fft[1]*fft[0]-4,
 		       SEEK_CUR);
             density=1;
             section=5;
@@ -618,6 +674,7 @@ void check_read(FILE* infile, struct unit_cell *c, struct contents *m,
       if (ion_force) reverse8n(ion_force,3*(*nsp)*(*nsp_max));
       if (ion_vel) reverse8n(ion_vel,3*(*nsp)*(*nsp_max));
       if (mag_mom) reverse8n(mag_mom,(*nsp)*(*nsp_max));
+      if (mag_dir) reverse8n(mag_dir,3*(*nsp)*(*nsp_max));
       if (species_charge) reverse8n(species_charge,*nsp);
     }
     /* The following two are both reversed when read in */
@@ -640,6 +697,8 @@ void check_read(FILE* infile, struct unit_cell *c, struct contents *m,
     kp->bs_n=*bs_nkpts;
     free(bs_wkpts);
     free(bs_kpts_pos);
+    free(bs_nkpts);
+    bs_nkpts=NULL;
     /* These are meaningful only if they differ from the SCF kpoints */
     if (kp->bs_n==kp->n){
       j=1;
@@ -726,25 +785,52 @@ void check_read(FILE* infile, struct unit_cell *c, struct contents *m,
             c->vol,conv);
 
   /* Rescue symmetry operations into sane structure */
-  if ((nsymops)&&(*nsymops>1)){
-    s->n=*nsymops;
-    s->ops=malloc(s->n*sizeof(struct sym_op));
-    if (!s->ops) error_exit("Malloc error for symops");
-    for(i=0;i<s->n;i++){
-      /* Translations are relative */
-      s->ops[i].tr=malloc(3*sizeof(double));
-      for(j=0;j<3;j++) s->ops[i].tr[j]=sym_disp[3*i]*c->basis[0][j]+
-			 sym_disp[3*i+1]*c->basis[1][j]+
-			 sym_disp[3*i+2]*c->basis[2][j];
-      /* But rotations are absolute */
-      for(j=0;j<3;j++)
-	for(k=0;k<3;k++)
-	  s->ops[i].mat[j][k]=sym_mat[9*i+3*j+k];
+  if ((nsymops)&&(sym_mat)&&(sym_disp)){
+    if (*nsymops>1){
+      s->n=*nsymops;
+      s->ops=malloc(s->n*sizeof(struct sym_op));
+      if (!s->ops) error_exit("Malloc error for symops");
+      for(i=0;i<s->n;i++){
+	/* Translations are relative */
+	s->ops[i].tr=malloc(3*sizeof(double));
+	/* And sometimes have fractional co-ords of 0.999... in place of 0 */
+	for(j=0;j<3;j++)
+	  if (aeq(fabs(sym_disp[3*i+j]),1.0)) sym_disp[3*i+j]=0;
+	for(j=0;j<3;j++) s->ops[i].tr[j]=sym_disp[3*i]*c->basis[0][j]+
+			   sym_disp[3*i+1]*c->basis[1][j]+
+			   sym_disp[3*i+2]*c->basis[2][j];
+	/* Rotations are absolute, but need transposing */
+	for(j=0;j<3;j++)
+	  for(k=0;k<3;k++)
+	    s->ops[i].mat[j][k]=sym_mat[9*i+j+3*k];
+      }
+      if (debug>1){
+	fprintf(stderr,"%d symmetry operations\n",s->n);
+	if (sym_mat) fprintf(stderr,"symmetry matrices read\n");
+      }
     }
     free(sym_disp);
+    sym_disp=NULL;
     free(sym_mat);
+    sym_mat=NULL;
   }
 
+  /* If spin directions present and are all z, discard */
+
+  if (mag_dir){
+    okay=1;
+    for(i=0;i<(*nsp)*(*nsp_max);i++){
+      if ((mag_dir[3*i]!=0)||(mag_dir[3*i+1]!=0)||(mag_dir[3*i+2]!=1)){
+	okay=0;
+	break;
+      }
+    }
+    if (okay){
+      free(mag_dir);
+      mag_dir=NULL;
+    }
+  }
+  
 /* Pack ions */
 
   m->n=0;
@@ -773,14 +859,14 @@ void check_read(FILE* infile, struct unit_cell *c, struct contents *m,
     }
     else lab=NULL;
     for(j=0;j<nionsp[i];j++){
-      if (mag_mom)
+      if (mag_dir)
+	for(k=0;k<3;k++)
+	  m->atoms[na].vspin[k]=mag_dir[3*(i*(*nsp_max)+j)+k]*
+	    mag_mom[i*(*nsp_max)+j];
+      else if (mag_mom)
         m->atoms[na].spin=mag_mom[i*(*nsp_max)+j];
-      else
-        m->atoms[na].spin=0;
       if (species_charge)
         m->atoms[na].chg=species_charge[i];
-      else
-        m->atoms[na].chg=0;
       m->atoms[na].atno=atsym2no(sym);
       m->atoms[na].label=lab;
       if((!m->atoms[na].atno)&&(debug))
@@ -794,7 +880,6 @@ void check_read(FILE* infile, struct unit_cell *c, struct contents *m,
         m->atoms[na].force[1]=ion_force[3*(i*(*nsp_max)+j)+1];  
         m->atoms[na].force[2]=ion_force[3*(i*(*nsp_max)+j)+2];  
       }
-      else for(k=0;k<3;k++)m->atoms[na].force[k]=0;
       /* Castep's units are Bohrs per Hartree time unit */
       /* Ours are Angstroms per picosecond */
       if (ion_vel){
@@ -802,11 +887,15 @@ void check_read(FILE* infile, struct unit_cell *c, struct contents *m,
         m->atoms[na].v[1]=ion_vel[3*(i*(*nsp_max)+j)+1]*BOHR/H_ps;
         m->atoms[na].v[2]=ion_vel[3*(i*(*nsp_max)+j)+2]*BOHR/H_ps;  
       }
-      else for(k=0;k<3;k++)m->atoms[na].v[k]=0;
       na++;
     }
   }
 
+  if (mag_dir){
+    free(mag_dir);
+    mag_dir=NULL;
+  }
+  
   /* Recreate a pseudopot block */
 
   smisc_size=0;
@@ -870,12 +959,6 @@ void check_read(FILE* infile, struct unit_cell *c, struct contents *m,
 /* Castep works in Bohr, we work in Angstroms */
   (*(s->tol))*=BOHR;
 
-  if ((debug>1)&&nsymops){  /* Note nsymops is not a valid pointer here */
-    fprintf(stderr,"%d symmetry operations\n",s->n);
-    if (sym_mat) fprintf(stderr,"symmetry matrices read\n");
-  }
-
-  
   /* Worry about dipole corrections */
 
   /* .check dip_corr_dir is always upper case, and anything on the
@@ -1013,10 +1096,10 @@ void reverse8n(double *data, int n){
 static void wave_read(FILE *infile, struct kpts *kp,
                       int gamma, struct grid *g, struct unit_cell *c,
                       struct es *elect, int *i_grid, struct contents *m){
-  int tmp,ns,b,i,k,nplwv,fft[3],dummy[5];
+  int tmp,junk[2],ns,b,i,k,nplwv,fft[3],dummy[5];
   int nsp,nspr;
   int *pwgrid=NULL,*pwg2=NULL;
-  long fpos;
+  long fpos,bytes_to_skip;
   double *dptr1;
   double kpoint[3];
   int k2;
@@ -1066,7 +1149,7 @@ static void wave_read(FILE *infile, struct kpts *kp,
   if (tmp!=12) error_exit("Error parsing wavefunction");
 
   fread(fft,12,1,infile);
-  fseek(infile,4,SEEK_CUR);
+  fread(junk,4,1,infile);
   if (endian) {
      reverse4(fft);
      reverse4(fft+1);
@@ -1139,11 +1222,14 @@ static void wave_read(FILE *infile, struct kpts *kp,
        	fseek(infile,24,SEEK_CUR);
 	fread(&nplwv,4,1,infile);
 	if (endian) reverse4(&nplwv);
-	fseek(infile,4,SEEK_CUR);
+	/* fseek(infile,4,SEEK_CUR);  - add to next fseek */
 	/* Skip three plane wave to grid component arrays */
-	fseek(infile,3*(8+4*(long)nplwv),SEEK_CUR);
+/*	fseek(infile,4+3*(8+4*(long)nplwv),SEEK_CUR); */
 	/* Skip nbands*nspinors complex wavefunction */
-	fseek(infile,nbands*elect->nspinors*(8+16*(long)nplwv),SEEK_CUR);
+/*	fseek(infile,nbands*elect->nspinors*(8+16*(long)nplwv),SEEK_CUR); */
+	/* Combine above two fseeks */
+	fseek(infile,4+3*(8+4*(long)nplwv)+
+		     nbands*elect->nspinors*(8+16*(long)nplwv),SEEK_CUR);
       }
     }
     occ_read(infile,elect,kp);
@@ -1155,7 +1241,9 @@ static void wave_read(FILE *infile, struct kpts *kp,
   for(ns=0;ns<elect->nbspins;ns++){
     for(k2=1;k2<=nkpts;k2++){
 /* record of kpoint[3],nplwv */
-      fseek(infile,4,SEEK_CUR);
+      fread(&tmp,4,1,infile);
+      if (endian) reverse4(&tmp);
+      if (tmp!=3*8+4) error_exit("Error parsing kpoint");
       fread(kpoint,3*8,1,infile);
       fread(&nplwv,4,1,infile);
       if (endian) reverse8n(kpoint,3);
@@ -1167,23 +1255,23 @@ static void wave_read(FILE *infile, struct kpts *kp,
         k=k2;
       else
         k=1+find_kpt(kpoint,kp);
-      if(debug>2) fprintf(stderr,"kpoint no %d (%f,%f,%f) nplwv=%d\n",
-                          k,kpoint[0],kpoint[1],kpoint[2],nplwv);
-      fseek(infile,4,SEEK_CUR);
+      if(debug>1) fprintf(stderr,
+		          "Spin=%d kpoint no %d (% f,% f,% f) nplwv=%d\n",
+                          ns,k,kpoint[0],kpoint[1],kpoint[2],nplwv);
+      fread(junk,4,1,infile);
 /* Read component to FFT grid mapping */
       fread(&tmp,4,1,infile);
       if (endian) reverse4(&tmp);
       if (4*nplwv!=tmp) error_exit("Error parsing band");
-      if (debug>2) fprintf(stderr,"Spin=%d kpt=%d nplwv=%d\n",ns,k,nplwv);
       if (inrange(k,kpt_range)&&
 	  ((elect->nspinors==2)||(inrange(ns,spin_range)))){
         if (!(pwgrid=malloc(12*nplwv))) error_exit("Malloc error for pwgrid");
         fread(pwgrid,tmp,1,infile);
-        fseek(infile,8,SEEK_CUR);
+	fread(junk,8,1,infile);
         fread(pwgrid+nplwv,tmp,1,infile);
-        fseek(infile,8,SEEK_CUR);
+	fread(junk,8,1,infile);
         fread(pwgrid+2*nplwv,tmp,1,infile);
-        fseek(infile,4,SEEK_CUR);
+	fread(junk,4,1,infile);
         if(endian) reverse4n(pwgrid,3*nplwv);
         if (debug>2) fprintf(stderr,"Read pwgrid\n");
 /* CASTEP stores these reciprocal space coeffs as -n/2 to n/2, whereas
@@ -1207,6 +1295,7 @@ static void wave_read(FILE *infile, struct kpts *kp,
 
       else fseek(infile,tmp*3+5*4,SEEK_CUR);
 
+      bytes_to_skip=0;
       for(b=1;b<=nbands;b++){
         if (debug>2) fprintf(stderr,"Start band %d\n",b);
         for(nspr=0;nspr<elect->nspinors;nspr++){
@@ -1217,13 +1306,18 @@ static void wave_read(FILE *infile, struct kpts *kp,
 	      inrange(b,band_range)){
 	    if (debug>2) fprintf(stderr,"starting band read,"
 				 " kpt %d, band %d, spin %d\n",k,b,nsp);
+	    if (bytes_to_skip){
+	      fseek(infile,bytes_to_skip,SEEK_CUR);
+	      bytes_to_skip=0;
+	    }
 	    fread(&tmp,4,1,infile);
 	    if (endian) reverse4(&tmp);
 	    if (tmp!=16*nplwv) error_exit("Error parsing wavefunction band");
 	    if (!(dptr1=malloc(16*nplwv)))
 	      error_exit("Malloc error in band read");
 	    fread(dptr1,16*nplwv,1,infile);
-	    fseek(infile,4,SEEK_CUR);
+	    /* fseek(infile,4,SEEK_CUR); */
+	    fread(junk,4,1,infile);
 	    if(endian) reverse8n(dptr1,2*nplwv);
 
 	    /* Care: ns loop starts at 0, k and b loops at 1 */
@@ -1234,10 +1328,14 @@ static void wave_read(FILE *infile, struct kpts *kp,
 	    free(dptr1); 
 	  }  /* if (inrange) */
 	  /* Consecutive fseeks in glibc are expensive, as each is an
-	   * lseek followed by a read(!). So we could optimise better here */
-	  else fseek(infile,16*nplwv+8,SEEK_CUR);
+	   * lseek followed by a read(!). So we optimise here */
+	  else bytes_to_skip+=16*nplwv+8;
 	} /*for(nspr=...) */
       } /* for(b=...) */
+      if (bytes_to_skip){
+	fseek(infile,bytes_to_skip,SEEK_CUR);
+	bytes_to_skip=0;
+      }
       if (pwgrid) {free(pwgrid); pwgrid=NULL;}
       if (pwg2) {free(pwg2); pwg2=NULL;}
     } /* for(k=...) */
@@ -1259,7 +1357,7 @@ static void wave_read(FILE *infile, struct kpts *kp,
 
 /* Note that the two spinor components share the same occupancy */
 void occ_read(FILE *infile, struct es *elect, struct kpts *kp){
-  int i,tmp,k,k2,ns,nbands,nspins,hit;
+  int i,tmp,junk[2],k,k2,ns,nbands,nspins,hit;
   double *dptr1,*dptr2;
   double kpoint[3];
 
@@ -1280,7 +1378,7 @@ void occ_read(FILE *infile, struct es *elect, struct kpts *kp){
     if (tmp!=24) error_exit("Error parsing end of wavefunction");
     fread(kpoint,3*8,1,infile);
     if (endian) reverse8n(kpoint,3);
-    fseek(infile,4,SEEK_CUR);
+    fread(junk,4,1,infile);
     /* Orbitals files may have the same kpt multiple times... */
     if (aeq(kp->kpts[k2].frac[0],kpoint[0])&&
           aeq(kp->kpts[k2].frac[1],kpoint[1])&&
@@ -1312,13 +1410,13 @@ void occ_read(FILE *infile, struct es *elect, struct kpts *kp){
       dptr2=elect->eval+nspins*nbands*k+ns*nbands;
       fread(dptr1,8*nbands,1,infile);
       if (endian) reverse8n(dptr1,nbands);
-      fseek(infile,4,SEEK_CUR);
+      fread(junk,4,1,infile);
       fread(&tmp,4,1,infile);
       if (endian) reverse4(&tmp);
       if (tmp!=8*nbands) error_exit("Error parsing end of wavefunction");
       fread(dptr2,8*nbands,1,infile);
       if (endian) reverse8n(dptr2,nbands);
-      fseek(infile,4,SEEK_CUR);
+      fread(junk,4,1,infile);
     }
   }
   if (elect->eval)
