@@ -10,7 +10,25 @@
  * in its keywords.
  */
 
-/* Copyright (c) 2018-2020 MJ Rutter 
+
+/* Note too Siesta's strange reading of include files
+ *
+ *  include foo.dat  -- entirely conventional, insert foo.dat here
+ *
+ * label < foo.dat -- parse foo.dat as fdf file, update only label,
+ * which may be a block...
+ *
+ * %block label < foo.dat
+ *
+ * parses as
+ *
+ * %block label
+ * [contents of foo.dat]
+ * [implicit %endblock label]
+ *
+ */
+
+/* Copyright (c) 2018-2023 MJ Rutter 
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -31,6 +49,7 @@
 #include<string.h>
 #include<ctype.h>
 #include<errno.h>
+#include<math.h>
 
 #include "c2xsf.h"
 
@@ -41,31 +60,38 @@
 static int fdfreadline(char *buffer, int len, char *dir,
                        struct infiles **filesp);
 static int fdfreadlength(char *buff, double *x);
+static int fdfreadlogical(char **ptr);
 static int fdfreadenergy(char *buff, double *x);
 static int fdfstrcasecmp(char **s1_in, char *s2);
 static int fdftokenmatch(char **s1_in, char *s2);
-static double siesta_pot_charge(char *prefix, char *label);
+static int fdfscanfile(char *filename, char *dirname, char *buffer,
+                       char *token, struct infiles **infile);
+static void fdf_unwind(struct infiles **infile);
+double siesta_pot_charge(char *prefix, char *label); /* Used by xv_read.c */
+void dequote(char **str);
 
 static char *ptr;
 
 void fdf_read(FILE* in, struct unit_cell *c, struct contents *m,
               struct kpts *kp, struct es *e){
-  int i,j,k,n,*atomspecies,mp_grid[3][3],okay;
+  int i,j,k,n,*atomspecies,mp_grid[3][3],okay,hit;
   int nspec,*species_to_atno;
-  double scale,shift[3],mp_off[3];
+  double scale,shift[3],mp_off[3],theta,phi;
   char *ptr2,*format;
   static char buffer[LINE_SIZE+1];
   double *vec,*abc,*atomcoords,*dmetol,latticeconstant,*atomspins;
   double *species_to_charge;
   char *siesta_misc[]={"ElectronicTemperature","MeshCutoff",
                        "MaxSCFIterations","MinSCFIterations",
-                       "OccupationFunction","SaveDeltaRho",
+                       "OccupationFunction","PAOBasisSize","PAOBasisType",
+		       "PAOSplitNorm","SaveDeltaRho",
                        "SaveElectrostaticPotential",
                        "SaveRho","SaveRhoXC",
                        "SaveTotalCharge","SaveTotalPotential",
                        "SlabDipoleCorrection","SolutionMethod",
-                       "WriteDM","XCauthors","XCfunctional",
-                       ""};
+                       "WriteDM","WriteMullikenPop","XCauthors",
+		       "XCfunctional",""};
+  char *siesta_block_misc[]={"PAOBasisAndPSlmax","XChybrid",""};
   char *spaces="                              ";  /* 30 spaces */
   char *prefix;
   char *dir;
@@ -82,7 +108,10 @@ void fdf_read(FILE* in, struct unit_cell *c, struct contents *m,
   files->f=in;
   files->next=files->last=NULL;
   files->name=NULL;
+  files->include=0;
   files->line=0;
+  files->ret=NULL;
+  files->count=0;
   
   m->n=0;
   nspec=0;
@@ -170,6 +199,23 @@ void fdf_read(FILE* in, struct unit_cell *c, struct contents *m,
       dict_add(m->dict,"Siesta_SystemLabel",ptr2);
     }
 
+    if (!fdftokenmatch(&ptr,"Spin")){
+      while(isspace(*ptr)) ptr++;
+      if (!strcasecmp(ptr,"polarized"))
+	e->nspins=2;
+      else if ((!strcasecmp(ptr,"non-colinear"))||
+	  (!strcasecmp(ptr,"non-collinear")))
+	e->nspinors=2;
+
+      continue;
+    }
+
+    if (!fdftokenmatch(&ptr,"SpinPolarized")){
+      if (fdfreadlogical(&ptr))
+	e->nspins=2;
+      continue;
+    }
+    
     for(i=0;siesta_misc[i][0];i++){
       if (!fdftokenmatch(&ptr,siesta_misc[i])){
         dict_strcat(m->dict,"Siesta_misc",siesta_misc[i]);
@@ -190,10 +236,11 @@ void fdf_read(FILE* in, struct unit_cell *c, struct contents *m,
     if (strncasecmp(ptr,"%block",6)) continue;
     ptr+=6;
 /* Eat leading spaces */
-    while(*ptr==' ') ptr++;
+    while(isspace(*ptr)) ptr++;
 /* Kill trailing spaces and things after # */
     ptr2=ptr;
-    while((*ptr2)&&(*ptr2!=' ')&&(*ptr2!='#')) ptr2++;
+    while((*ptr2)&&(*ptr2!=' ')&&(*ptr2!='#')&&(*ptr2!='!')&&(*ptr2!=';'))
+      ptr2++;
     *ptr2=0; /* Either it was a null, or it should be one... */
 
     if (debug>2) fprintf(stderr,"Found a %s block\n",ptr);
@@ -232,6 +279,12 @@ void fdf_read(FILE* in, struct unit_cell *c, struct contents *m,
       }
       fdfreadline(buffer,LINE_SIZE,dir,&files);
     }
+    else if (!fdfstrcasecmp(&ptr,"AtomicCoordinatesOrigin")){
+      fdfreadline(buffer,LINE_SIZE,dir,&files);
+      if (sscanf(buffer,"%lf %lf %lf",shift,shift+1,shift+2)!=3)
+        error_exit("Error parsing atomic coordinates origin");
+      fdfreadline(buffer,LINE_SIZE,dir,&files);
+    }
     else if (!fdfstrcasecmp(&ptr,"ChemicalSpeciesLabel")){
       if (!nspec)
         error_exit("ChemicalSpeciesLabel preceeds NumberOfSpecies");
@@ -261,8 +314,8 @@ void fdf_read(FILE* in, struct unit_cell *c, struct contents *m,
     else if (!fdfstrcasecmp(&ptr,"DMInitSpin")){
       if (m->n==0)
         error_exit("DMInitSpin preceeds NumberOfAtoms");
-      atomspins=malloc(m->n*sizeof(double));
-      for(i=0;i<m->n;i++) atomspins[i]=0;
+      atomspins=malloc(3*m->n*sizeof(double));
+      for(i=0;i<3*m->n;i++) atomspins[i]=0;
       while(1){
         fdfreadline(buffer,LINE_SIZE,dir,&files);
         i=sscanf(buffer,"%d%n",&j,&k);
@@ -272,13 +325,14 @@ void fdf_read(FILE* in, struct unit_cell *c, struct contents *m,
         ptr=buffer+k;
         while (*ptr==' ') ptr++;
         if (!tokenmatch(&ptr,"+")){
-          atomspins[j-1]=1;
+          atomspins[3*(j-1)]=1;
         }
         else if (!tokenmatch(&ptr,"-")){
-          atomspins[j-1]=-1;
+          atomspins[3*(j-1)]=-1;
         }
         else{
-          if (sscanf(ptr,"%lf",atomspins+j-1)!=1)
+          if (sscanf(ptr,"%lf %lf %lf",atomspins+3*(j-1),
+		     atomspins+3*(j-1)+1,atomspins+3*(j-1)+2)<1)
             error_exit("Error parsing spin");
         }
       }
@@ -319,10 +373,28 @@ void fdf_read(FILE* in, struct unit_cell *c, struct contents *m,
       }
     }
     else{
-      fprintf(stderr,"Ignoring %s\n",ptr);
-      while(fdfreadline(buffer,LINE_SIZE,dir,&files)){
-        ptr=buffer;
-        if (!strncasecmp(ptr,"%endblock",9)) break;
+      hit=0;
+      for(i=0;siesta_block_misc[i][0];i++){
+	if (!fdftokenmatch(&ptr,siesta_block_misc[i])){
+	  hit=1;
+	  dict_strcat(m->dict,"Siesta_misc","\n");
+	  dict_strcat(m->dict,"Siesta_misc",buffer);
+	  dict_strcat(m->dict,"Siesta_misc","\n");
+	  while(fdfreadline(buffer,LINE_SIZE,dir,&files)){
+	    dict_strcat(m->dict,"Siesta_misc",buffer);
+	    dict_strcat(m->dict,"Siesta_misc","\n");
+	    if (!strncasecmp(buffer,"%endblock",9)) break;
+	  }
+	  dict_strcat(m->dict,"Siesta_misc","\n");
+	  break;
+	}
+      }
+      if (!hit){
+	fprintf(stderr,"Ignoring %s\n",ptr);
+	while(fdfreadline(buffer,LINE_SIZE,dir,&files)){
+	  ptr=buffer;
+	  if (!strncasecmp(ptr,"%endblock",9)) break;
+	}
       }
     }
   }
@@ -416,8 +488,25 @@ void fdf_read(FILE* in, struct unit_cell *c, struct contents *m,
   /* Add spins */
   
   if (atomspins){
-    for(i=0;i<m->n;i++)
-      m->atoms[i].spin=atomspins[i];
+    okay=0;
+    for(i=0;i<m->n;i++){
+      if ((atomspins[3*i+1]!=0)||(atomspins[3*i+2]!=0)){
+	okay=1;
+	break;
+      }
+    }
+    if (okay){
+      for(i=0;i<m->n;i++){
+	theta=M_PI*atomspins[3*i+1]/180;
+	phi=M_PI*atomspins[3*i+2]/180;
+	m->atoms[i].vspin[0]=atomspins[3*i]*sin(theta)*cos(phi);
+	m->atoms[i].vspin[1]=atomspins[3*i]*sin(theta)*sin(phi);
+	m->atoms[i].vspin[2]=atomspins[3*i]*cos(theta);
+      }
+    }
+    else
+      for(i=0;i<m->n;i++)
+	m->atoms[i].spin=atomspins[3*i];
     free(atomspins);
   }
   
@@ -427,11 +516,14 @@ void fdf_read(FILE* in, struct unit_cell *c, struct contents *m,
     e->etol=(*dmetol)/m->n;
     free(dmetol);
   }
+
+  free(files);
+  
 }
 
 int fdfreadline(char *buffer, int len, char *dir, struct infiles **filesp){
-  int off,success2,inc;
-  char *ptr,*success,*p2;
+  int i,off,success2,inc;
+  char *ptr,*success,*p2,token[LINE_SIZE];
   
   while((success=fgets(buffer,len,(*filesp)->f))){ /* fgets() always
                                                     null terminates,
@@ -445,9 +537,9 @@ int fdfreadline(char *buffer, int len, char *dir, struct infiles **filesp){
 
 /* Eat leading spaces */
     ptr=buffer;
-    while(*ptr==' ') ptr++;
+    while(isspace(*ptr)) ptr++;
 /* Skip comments and blank lines */
-    if ((*ptr=='#')||(*ptr==0)) continue;
+    if ((*ptr=='#')||(*ptr=='!')||(*ptr==';')||(*ptr==0)) continue;
     break;
   }
 
@@ -470,12 +562,21 @@ int fdfreadline(char *buffer, int len, char *dir, struct infiles **filesp){
     if (!success2) return(0);
   }
 
+  /* Shift buffer to remove leading whitespace */
   off=ptr-buffer;
   if (off){
     while(*ptr) {*(ptr-off)=*ptr;ptr++;}
     *(ptr-off)=0;
   }
 
+  ptr=buffer;
+  
+  /* return read %endblock line, but reset files to old value */
+  if (((*filesp)->include==2)&&(!strncasecmp(ptr,"%endblock",9))){
+    fdf_unwind(filesp);
+    return 1;
+  }
+
   if (!tokenmatch(&ptr,"%include")){
     while (*ptr==' ') ptr++;
     p2=ptr;
@@ -497,13 +598,22 @@ int fdfreadline(char *buffer, int len, char *dir, struct infiles **filesp){
 
   /* do we now have a < ? */
   if (*p2=='<'){
+    for(i=0;!(isspace(buffer[i]));i++)
+      token[i]=buffer[i];
+    token[i]=0;
     ptr=p2+1;
     while (*ptr==' ') ptr++;
     p2=ptr;
     while ((*p2)&&(*p2!=' ')) p2++;
     *p2=0;
-    include_file(filesp,dir,ptr);
-    return fdfreadline(buffer,LINE_SIZE,dir,filesp);
+
+    i=fdfscanfile(ptr,dir,buffer,token,filesp);
+    if (!i){
+      fprintf(stderr,"Warning: %s not found in < included file %s\n",
+	      token,ptr);
+      return fdfreadline(buffer,LINE_SIZE,dir,filesp);
+    }
+    else return 1;
   }
     
   
@@ -530,9 +640,13 @@ int fdfreadline(char *buffer, int len, char *dir, struct infiles **filesp){
 
 void include_file(struct infiles **file, char *dir, char *ptr){
   struct infiles *fs;
-  char *name;
+  char *name,*infile;
 
   name=NULL;
+  infile=ptr;
+  dequote(&infile);
+  ptr=infile;
+  
   fs=*file;
   fs->next=malloc(sizeof(struct infiles));
   if (!fs->next) error_exit("Malloc error for files_next in cell_read");
@@ -540,6 +654,10 @@ void include_file(struct infiles **file, char *dir, char *ptr){
   fs->next->next=NULL;
   fs=fs->next;
   fs->f=NULL;
+  fs->count=(*file)->count+1;
+  if (fs->count>30)
+    error_exit("maximum include depth of 30 exceeded -- infinite recursion?");
+  
   if (dir){
     name=malloc(strlen(dir)+strlen(ptr)+1);
     if (!name) error_exit("malloc error for filename");
@@ -558,6 +676,7 @@ void include_file(struct infiles **file, char *dir, char *ptr){
     fprintf(stderr,"Error: failed to open include file ");
     if (name) fprintf(stderr,"%s or",name);
     fprintf(stderr,"%s\n",ptr);
+    
   }
   if (debug)
     fprintf(stderr,"Opened included file %s\n",fs->name);
@@ -565,6 +684,10 @@ void include_file(struct infiles **file, char *dir, char *ptr){
   fs->line=0;
   fs->include=1;  /* This gets set to zero if a block include which
                      * needs an %endblock added */   
+  if ((*file)->include)
+  fs->include=(*file)->include;
+  fs->count=(*file)->count+1;
+  fs->ret=(*file)->ret;
   *file=fs;
 }
 
@@ -682,6 +805,8 @@ static int fdfstrcasecmp(char **s1_in, char *s2){
   while((*s1)&&(*s2)){
     if ((*s1=='_')||(*s1=='-')||(*s1=='.'))
       s1++;
+    else if ((*s2=='_')||(*s2=='-')||(*s2=='.'))
+      s2++;
     else if (toupper(*s1)==toupper(*s2)){
       s1++;
       s2++;
@@ -716,23 +841,35 @@ static int fdftokenmatch(char **s1_in, char *s2){
   return 1;
 }
 
-static double siesta_pot_charge(char *prefix, char *label){
+/* Read charge from .psf file */
+double siesta_pot_charge(char *prefix, char *label){
   char *name;
   int i,idummy;
   double dummy,charge;
   FILE *pseudo;
   char buffer[LINE_SIZE+1];
 
-  name=malloc(strlen(prefix)+strlen(label)+5);
+  if (prefix)
+    name=malloc(strlen(prefix)+strlen(label)+10);
+  else
+    name=malloc(strlen(label)+10);
+
   name[0]=0;
-    
-  strcpy(name,prefix);
+  if (prefix) strcpy(name,prefix);
   strcat(name,label);
   strcat(name,".psf");
 
   pseudo=fopen(name,"rb");
 
   if (!pseudo) {
+    name[0]=0;
+    if (prefix) strcpy(name,prefix);
+    strcat(name,label);
+    strcat(name,".out.psf");
+    pseudo=fopen(name,"rb");
+  }
+
+  if (!pseudo){
     fprintf(stderr,"Failed to open %s\n",name);
     free(name);
     return 0;
@@ -748,12 +885,136 @@ static double siesta_pot_charge(char *prefix, char *label){
     fprintf(stderr,"Failed to read charge from %s\n",name);
     charge=0;
   }
-
+  else if (debug)
+    fprintf(stderr,"Label %s charge %lf read from %s\n",
+	    label,charge,name);
+    
   fclose(pseudo);
   free(name);
 
-  if (debug>1) fprintf(stderr,"Label %s charge %lf\n",label,charge);
-  
   return charge;
+}
+
+/* scan file(s) for a given token in response to "<" syntax */
+
+static int fdfscanfile(char *filename, char *dirname, char *buffer,
+                       char *token, struct infiles **infile){
+  char *ptr;
+  int found,block;
+  struct infiles *files;
+
+  if (debug>2)
+    fprintf(stderr,"fdfscanfile for %s in %s called\n",
+	    token,filename);
+
+  block=found=0;
+
+  files=*infile;
+  include_file(infile,dirname,filename);
+  (*infile)->ret=files;
+  if ((*infile)->f==NULL){
+    free((*infile)->name);
+    free(*infile);
+    *infile=files;
+    return 0;
+  }
   
+  while(fdfreadline(buffer,LINE_SIZE,dirname,infile)){
+    ptr=buffer;
+    block=0;
+    if (!strncasecmp(ptr,"%block",6)){
+      block=1;
+      ptr+=6;
+      while(isspace(*ptr)) ptr++;
+    }
+    if (!fdftokenmatch(&ptr,token)){
+      found=1;
+      break;
+    }
+  }
+
+  if (debug>1)
+    fprintf(stderr,"fdfscanfile for %s in %s returns %d block=%d %s\n",
+	    token,(*infile)->last->name,found,block,buffer);
+
+  if (block&&found){
+    (*infile)->include=2;
+  }
+  else{
+    fdf_unwind(infile);
+  }
+    
+  return found;
+}
+
+/* Deal with potentially multiple back-outs caused by Siesta's "<" syntax */
+static void fdf_unwind(struct infiles **infile){
+  struct infiles *target;
+
+  target=(*infile)->ret;
+
+  if (!target) error_exit("asked to unwind to null in fdf_unwind");
+  
+  while(*infile!=target){
+    fclose((*infile)->f);
+    free((*infile)->name);
+    *infile=(*infile)->last;
+    free((*infile)->next);
+  }
+}
+
+void dequote(char **str){
+  char *p;
+
+  p=*str;
+  
+  if (((*p)=='"')&&(strchr(p+1,'"'))){
+    p++;
+    *(strchr(p,'"'))=0;
+    *str=p;
+    return;
+  }
+
+  if (((*p)=='\'')&&(strchr(p+1,'\''))){
+    p++;
+    *(strchr(p,'\''))=0;
+    *str=p;
+    return;
+  }
+}
+
+static int fdfreadlogical(char **ptr){
+  
+  while((**ptr)&&(isspace(**ptr))) (*ptr)++;
+  if (!(**ptr)) return 1;
+  if (**ptr=='#'){
+    while (**ptr) (*ptr)++;
+    return 1;
+  }
+  
+  if (!tokenmatch(ptr,".true."))
+    return 1;
+
+  if (!tokenmatch(ptr,"true"))
+    return 1;
+
+  if (!tokenmatch(ptr,"yes"))
+    return 1;
+
+  if (!tokenmatch(ptr,"T"))
+    return 1;
+
+  if (!tokenmatch(ptr,".false."))
+    return 0;
+
+  if (!tokenmatch(ptr,"false"))
+    return 0;
+
+  if (!tokenmatch(ptr,"no"))
+    return 0;
+
+  if (!tokenmatch(ptr,"F"))
+    return 0;
+
+  return 0;
 }
