@@ -43,6 +43,8 @@ static int qe_readline(char **p, FILE* infile);
 static int qe_int_read(char **ptr,int *n){
   int i;
   
+  if (!*ptr) return 0;
+
   if ((sscanf(*ptr,"%d%n",n,&i)!=1)&&(sscanf(*ptr," = %d%n",n,&i)!=1))
     return 0;
 
@@ -56,6 +58,8 @@ static int qe_int_read(char **ptr,int *n){
 static int qe_float_read(char **ptr,double *x){
   int i,j,ok;
   
+  if (!*ptr) return 0;
+
   if ((sscanf(*ptr,"%lf%n",x,&i)!=1)&&(sscanf(*ptr," = %lf%n",x,&i)!=1))
     return 0;
 
@@ -67,7 +71,7 @@ static int qe_float_read(char **ptr,double *x){
       if (sscanf(*ptr,"%d%n",&j,&i)==1)
         *x*=pow(10.0,(double)j);
       else
-        *ptr-=1;
+        *ptr-=1+i;
     }
   }
   
@@ -80,7 +84,9 @@ static int qe_float_read(char **ptr,double *x){
 
 static int qe_logical_read(char **ptr,int *l){
 
-  while((*ptr)&&(isspace(**ptr))) (*ptr)++;
+  if (!*ptr) return 0;
+
+  while((**ptr)&&(isspace(**ptr))) (*ptr)++;
 
   if (!strncasecmp(*ptr,".true.",6)){
     *ptr+=6;
@@ -103,6 +109,8 @@ static int qe_string_read(char **ptr, char **str){
   int quote;
   char *p1,*p2;
   
+  if (!*ptr) return 0;
+
   while ((**ptr)&&(isspace(**ptr))) (*ptr)++;
 
   quote=0;
@@ -129,7 +137,7 @@ void qe_read(FILE* infile, struct unit_cell *c, struct contents *m,
   double celldm[7];
   double *abc,x,y,z,c2,alpha,beta,gamma,scale;
   double tx,ty,tz,u,v,dtmp;
-  double *magmom;
+  double *magmom,*angle1,*angle2;
   char sym[4],*ptr,*p2,*p3;
   char *aspec;
   int atoms_are_abs,off[3];
@@ -143,7 +151,7 @@ void qe_read(FILE* infile, struct unit_cell *c, struct contents *m,
   namelist=0;
   abc=NULL;
   aspec=NULL;
-  magmom=NULL;
+  magmom=angle1=angle2=NULL;
   atoms_are_abs=0;
   for(i=0;i<7;i++)celldm[i]=0;
   ptr="";
@@ -445,6 +453,40 @@ void qe_read(FILE* infile, struct unit_cell *c, struct contents *m,
 	if (!qe_float_read(&ptr,magmom+i))
           error_exit("Error parsing starting_magnetisation");
       }
+      else if (!strncasecmp(ptr,"angle1(",
+			    strlen("angle1("))){
+	if (!ntyp) error_exit("angle1 before ntyp");
+	if (!angle1) {
+	  angle1=malloc(ntyp*sizeof(double));
+	  if (!angle1) error_exit("malloc error for angle1");
+	  for(j=0;j<ntyp;j++) angle1[j]=0;
+	}
+	ptr+=strlen("angle1(");
+	if (sscanf(ptr,"%d",&i)!=1)
+          error_exit("Error parsing angle1");
+	i--;
+        ptr+=2;
+	if (!qe_float_read(&ptr,angle1+i))
+          error_exit("Error parsing angle1");
+	angle1[i]*=M_PI/180;
+      }
+      else if (!strncasecmp(ptr,"angle2(",
+			    strlen("angle2("))){
+	if (!ntyp) error_exit("angle2 before ntyp");
+	if (!angle2) {
+	  angle2=malloc(ntyp*sizeof(double));
+	  if (!angle2) error_exit("malloc error for angle2");
+	  for(j=0;j<ntyp;j++) angle2[j]=0;
+	}
+	ptr+=strlen("angle2(");
+	if (sscanf(ptr,"%d",&i)!=1)
+          error_exit("Error parsing angle2");
+	i--;
+        ptr+=2;
+	if (!qe_float_read(&ptr,angle2+i))
+          error_exit("Error parsing angle2");
+	angle2[i]*=M_PI/180;
+      }
       else{
         while ((*ptr)&&(!isspace(*ptr))) ptr++;
       }
@@ -915,9 +957,19 @@ void qe_read(FILE* infile, struct unit_cell *c, struct contents *m,
       if (!ptr) ptr=atno2sym(m->atoms[i].atno);
       for(j=0;j<ntyp;j++)
 	if (!strcasecmp(ptr,aspec+4*j)) break;
-      if (!strcasecmp(ptr,aspec+4*j))
-	m->atoms[i].spin=magmom[j];
+      if (!strcasecmp(ptr,aspec+4*j)){
+	if ((angle1)&&(angle2)){
+	  m->atoms[i].vspin[0]=magmom[j]*sin(angle1[j])*cos(angle2[j]);
+	  m->atoms[i].vspin[1]=magmom[j]*sin(angle1[j])*sin(angle2[j]);
+	  m->atoms[i].vspin[2]=magmom[j]*cos(angle1[j]);
+	}
+	else
+	  m->atoms[i].spin=magmom[j];
+      }
     }
+    if (angle2) free(angle2);
+    if (angle1) free(angle1);
+    free(magmom);
   }
 
   if (aspec) free(aspec);
