@@ -3,8 +3,11 @@
 /* This should be a proper XML parser. It isn't */
 
 /* Energies here are in Hartrees, but in a .in file are in Rydbergs?! */
+/* Angles here are in radians, but in a .in file are in degrees */
 /* Lengths are in Bohr */
 
+/* 1/2024 -- add QE 7.x support */
+
 #include<stdio.h>
 #include<stdlib.h> /* malloc */
 #include<string.h>
@@ -19,6 +22,8 @@
 static void read_step(FILE *infile, char *buffer, struct time_series *ts);
 static void read_basis(FILE *infile, struct unit_cell *c);
 static void read_atoms(FILE *infile, struct contents *m);
+static int qe_read_xml_floats(char *line, char *key, double *data,
+			       int n, FILE* infile);
 
 void qe_rho_read(FILE* infile, struct unit_cell *c, struct contents *m,
 		 struct kpts *k, struct symmetry *s, struct grid *g,
@@ -40,10 +45,10 @@ void qe_xml_read(FILE* infile, char *filename, struct unit_cell *c,
   char buffer[LINE_SIZE+1];
   char key[6];
   char *cptr,*cptr2,*cptr3,*den_name,*den_name2,*cwd;
-  int i,j,jj,tmp;
+  int i,j,jj,tmp,vspin;
   FILE *ch,*pot;
   struct sp {char *name; double mass; char *p_pot;
-    double chg; double mag;} *species;
+    double chg; double mag; double mtheta; double mphi;} *species;
   int nspec,nsym,fft[3],*iptr,nkpt,nbands,ver_maj,ver_min;
   double *symrel,*symtr,mat[3][3],*dptr,alat,sum;
   struct kpts *k_in;
@@ -86,6 +91,8 @@ void qe_xml_read(FILE* infile, char *filename, struct unit_cell *c,
   if (i==EOF) error_exit("Input file contains no lines!");
 
   fgets(buffer,LINE_SIZE,infile);
+  /* Consume comment lines */
+  while ((strstr(buffer,"<!--")==buffer)&&(fgets(buffer,LINE_SIZE,infile)));
 
   if (!strstr(buffer,"quantum-espresso"))
     error_exit("'quantum-espresso' not found on 2nd line of XML file");
@@ -154,12 +161,17 @@ void qe_xml_read(FILE* infile, char *filename, struct unit_cell *c,
     else if (strstr(buffer,"<forc_conv_thr>")){
       cptr=strstr(buffer,"<forc_conv_thr>")+15;
       dptr=malloc(sizeof(double));
+      if (!dptr) error_exit("Malloc error for forc_conv_thr");
       i=sscanf(cptr,"%lf",dptr);
-      if (i) dict_add(m->dict,"QE_forc_conv_thr",dptr);
+      if (i){
+	(*dptr)*=2; /* Ha in xml, but Ry in input files */
+	dict_add(m->dict,"QE_forc_conv_thr",dptr);
+      }
     }
     else if (strstr(buffer,"<tot_charge>")){
       cptr=strstr(buffer,"<tot_charge>")+12;
       e->charge=malloc(sizeof(double));
+      if (!e->charge) error_exit("Malloc error for e->charge");
       i=sscanf(cptr,"%lf",e->charge);
       if (!i) {
         fprintf(stderr,"Warning: failed to parse <tot_charge>\n");
@@ -180,6 +192,7 @@ void qe_xml_read(FILE* infile, char *filename, struct unit_cell *c,
       cptr=strstr(buffer,"<nosym>")+7;
       if (!strncmp(cptr,"true",4)){ /* record only if not the default */
 	iptr=malloc(sizeof(int));
+	if (!iptr) error_exit("Malloc error for nosym");
 	*iptr=1;
 	dict_add(m->dict,"QE_nosym",iptr);
       }
@@ -188,8 +201,13 @@ void qe_xml_read(FILE* infile, char *filename, struct unit_cell *c,
   
   /* Now scan for output section, also reading timesteps if req */
 
+  /* Note QE 6.x uses <step n_step="1"> on a line of its own at the start
+     of each step, and thus helpfully numbers all the steps.
+     QE 7.x uses simply <step> and there is no numbering in the file,
+     which may upset Humans trying to extract things manually. */
+  
   while(fgets(buffer,LINE_SIZE,infile)){
-    if ((ts)&&(strstr(buffer,"<step n_step=")))
+    if ((ts)&&((strstr(buffer,"<step n_step="))||(strstr(buffer,"<step>"))))
       read_step(infile,buffer,ts);
     if (strstr(buffer,"<output>")) break;
   }
@@ -204,8 +222,8 @@ void qe_xml_read(FILE* infile, char *filename, struct unit_cell *c,
       read_basis(infile,c);
     }
     else if(strstr(buffer,"<atomic_structure ")){
-      cptr=strstr(buffer,"<atomic_structure>");
-      cptr=strstr(buffer,"alat=");
+      cptr=strstr(buffer,"<atomic_structure ");
+      cptr=strstr(cptr,"alat=");
       cptr+=5;
       if (*cptr=='"') cptr++;
       sscanf(cptr,"%lf",&alat);
@@ -219,6 +237,7 @@ void qe_xml_read(FILE* infile, char *filename, struct unit_cell *c,
       if (cptr) {
 	sscanf(cptr+4,"%d",&k_in->n);
 	k_in->kpts=malloc(k_in->n*sizeof(struct atom));
+	init_atoms(k_in->kpts,k_in->n);
 	if (!k_in->kpts) error_exit("Malloc error for kpts");
 	for(i=0;i<k_in->n;i++){
 	  fgets(buffer,LINE_SIZE,infile);
@@ -230,12 +249,8 @@ void qe_xml_read(FILE* infile, char *filename, struct unit_cell *c,
 	  cptr++;
 	  j=sscanf(cptr,"%lf",&k_in->kpts[i].wt);
 	  if (j!=1) error_exit("Error parsing kpt weight");
-	  while((*cptr)&&(*cptr!='>')) cptr++;
-	  if (!*cptr) error_exit("Kpt misparse 1");
-	  cptr++;
-	  j=sscanf(cptr,"%lf %lf %lf",k_in->kpts[i].abs,
-		   k_in->kpts[i].abs+1,k_in->kpts[i].abs+2);
-	  if (j!=3) error_exit("Kpt misparse 2");
+	  if (qe_read_xml_floats(buffer,"<k_point",k_in->kpts[i].abs,3,infile)
+	      !=3) error_exit("Error parsing kpt");
 	}
       }
       else{
@@ -308,12 +323,8 @@ void qe_xml_read(FILE* infile, char *filename, struct unit_cell *c,
 	  cptr++;
 	  j=sscanf(cptr,"%lf",&k->kpts[nkpt].wt);
 	  if (j!=1) error_exit("Error parsing kpt weight");
-	  while((*cptr)&&(*cptr!='>')) cptr++;
-	  if (!*cptr) error_exit("Kpt misparse 1");
-	  cptr++;
-	  j=sscanf(cptr,"%lf %lf %lf",k->kpts[nkpt].abs,
-		   k->kpts[nkpt].abs+1,k->kpts[nkpt].abs+2);
-	  if (j!=3) error_exit("Kpt misparse 2");
+	  if (qe_read_xml_floats(buffer,"<k_point",k->kpts[nkpt].abs,3,infile)
+	      !=3) error_exit("Error parsing kpt");
 	}
 	if (strstr(buffer,"<occupations ")){
 	  cptr=strstr(buffer,"size=");
@@ -328,19 +339,10 @@ void qe_xml_read(FILE* infile, char *filename, struct unit_cell *c,
 	    if (!e->occ)
               e->occ=malloc(nbands*e->nspins*k->n*sizeof(double));
 	    if (!e->occ) error_exit("Malloc error for occupations");
-	    j=nbands*e->nspins*nkpt;
-	    while((*cptr)&&(*cptr!='>')) cptr++;
-	    i=0;
-	    while(i<nbands*e->nspins){
-	      while(sscanf(cptr,"%lf%n",e->occ+j+i,&tmp)==1){
-		cptr+=tmp;
-		i++;
-	      }
-	      if (i<nbands*e->nspins){
-		fgets(buffer,LINE_SIZE,infile);
-		cptr=buffer;
-	      }
-	    }
+	    if (qe_read_xml_floats(buffer,"<occupations",
+				   e->occ+nbands*e->nspins*nkpt,
+				   nbands*e->nspins,infile)!=nbands*e->nspins)
+	      error_exit("Error reading occupations");
 	  }
 	  else
 	    fprintf(stderr,"Unexpected size in occupations. Ignoring\n");
@@ -357,19 +359,10 @@ void qe_xml_read(FILE* infile, char *filename, struct unit_cell *c,
 	  if (tmp==nbands*e->nspins){
 	    if (!e->eval) e->eval=malloc(nbands*e->nspins*k->n*sizeof(double));
 	    if (!e->eval) error_exit("Malloc error for evals");
-	    j=nbands*e->nspins*nkpt;
-	    while((*cptr)&&(*cptr!='>')) cptr++;
-	    i=0;
-	    while(i<nbands*e->nspins){
-	      while(sscanf(cptr,"%lf%n",e->eval+j+i,&tmp)==1){
-		cptr+=tmp;
-		i++;
-	      }
-	      if (i<nbands*e->nspins){
-		fgets(buffer,LINE_SIZE,infile);
-		cptr=buffer;
-	      }
-	    }
+	    if (qe_read_xml_floats(buffer,"<eigenvalues",
+				   e->eval+nbands*e->nspins*nkpt,
+				   nbands*e->nspins,infile)!=nbands*e->nspins)
+	      error_exit("Error reading eigenvalues");
 	  }
 	  else
 	    fprintf(stderr,"Unexpected size in eigenvalues. Ignoring\n");
@@ -378,13 +371,13 @@ void qe_xml_read(FILE* infile, char *filename, struct unit_cell *c,
       nkpt++;
 
     }
-    else if(strstr(buffer,"<lsda>true</lsda>")){
+    else if (strstr(buffer,"<lsda>true</lsda>")){
       e->nspins=e->nbspins=2;
     }
-    else if(strstr(buffer,"<noncolin>true</noncolin>")){
+    else if (strstr(buffer,"<noncolin>true</noncolin>")){
       e->nspinors=2;
     }
-    else if(strstr(buffer,"<forces ")){
+    else if (strstr(buffer,"<forces ")){
       for(i=0;i<m->n;i++){
         fgets(buffer,LINE_SIZE,infile);
         j=sscanf(buffer,"%lf %lf %lf",m->atoms[i].force,m->atoms[i].force+1,
@@ -406,9 +399,13 @@ void qe_xml_read(FILE* infile, char *filename, struct unit_cell *c,
       if (i!=1) error_exit("Error parsing atomic_species nspec\n%s");
       if (debug>2) fprintf(stderr,"nspec=%d\n",nspec);
       species=malloc(nspec*sizeof(struct sp));
+      if (!species) error_exit("malloc error for species");
       for(i=0;i<nspec;i++){
         species[i].mass=0;
+	species[i].chg=0;
         species[i].mag=0;
+        species[i].mtheta=0;
+        species[i].mphi=0;
         species[i].name=NULL;
         species[i].p_pot=NULL;
         while(fgets(buffer,LINE_SIZE,infile))
@@ -421,6 +418,7 @@ void qe_xml_read(FILE* infile, char *filename, struct unit_cell *c,
         if (*cptr2!='"') error_exit("Error parsing species name");
         *cptr2=0;
         species[i].name=malloc((cptr2-cptr)+1);
+	if (!species[i].name) error_exit("malloc error for species name");
         strcpy(species[i].name,cptr);
         while(fgets(buffer,LINE_SIZE,infile)){
           if (strstr(buffer,"</species>")) break;
@@ -436,6 +434,8 @@ void qe_xml_read(FILE* infile, char *filename, struct unit_cell *c,
               fprintf(stderr,"Warning, error parsing pseudo_file\n");
             else{
               species[i].p_pot=malloc((cptr2-cptr)+1);
+	      if (!species[i].p_pot)
+		error_exit("malloc error for species pot");
               strncpy(species[i].p_pot,cptr,cptr2-cptr);
               species[i].p_pot[cptr2-cptr]=0;
             }
@@ -448,6 +448,22 @@ void qe_xml_read(FILE* infile, char *filename, struct unit_cell *c,
 	      fprintf(stderr,
 		      "Warning, error parsing species magnetic moment\n");
           }
+	  else if (strstr(buffer,"<spin_teta>")){
+	    cptr=strstr(buffer,"<spin_teta>")+
+	      strlen("<spin_teta>");
+	    j=sscanf(cptr,"%lf",&species[i].mtheta);
+            if (j!=1)
+	      fprintf(stderr,
+		      "Warning, error parsing species magnetic angle\n");
+          }
+	  else if (strstr(buffer,"<spin_phi>")){
+	    cptr=strstr(buffer,"<spin_phi>")+
+	      strlen("<spin_phi>");
+	    j=sscanf(cptr,"%lf",&species[i].mphi);
+            if (j!=1)
+	      fprintf(stderr,
+		      "Warning, error parsing species magnetic angle\n");
+          }
         }
       }
     }
@@ -475,37 +491,13 @@ void qe_xml_read(FILE* infile, char *filename, struct unit_cell *c,
             if(!strstr(buffer,"crystal_symmetry")){i--;continue;}
             while(fgets(buffer,LINE_SIZE,infile))
               if(strstr(buffer,"<rotation ")) break;
-            if (!fgets(buffer,LINE_SIZE,infile))
-              error_exit("Unexpected EOF reading symmetries");
-            cptr=buffer;
-            j=0;
-            do{
-              while (sscanf(cptr,"%lf%n",symrel+9*i+j,&jj)>0){
-                cptr+=jj;
-                j++;
-              }
-              if (!fgets(buffer,LINE_SIZE,infile))
-                error_exit("Error reading symmetry operations");
-              cptr=buffer;
-            } while (j<9);
+	    if (qe_read_xml_floats(buffer,"<rotation",symrel+9*i,9,infile)!=9)
+	      error_exit("Error reading rotation matrix");
             while(fgets(buffer,LINE_SIZE,infile))
               if(strstr(buffer,"<fractional_translation>")) break;
-            cptr=strstr(buffer,"<fractional_translation>");
-            if (!cptr) error_exit("Error reading symmetry operations");
-            cptr+=strlen("<fractional_translation>");
-            s->ops[i].tr=malloc(3*sizeof(double));
-            if (!s->ops[i].tr)
-              error_exit("Malloc error for symmetry translation");
-            j=0;
-            do{
-              while (sscanf(cptr,"%lf%n",symtr+3*i+j,&jj)>0){
-                cptr+=jj;
-                j++;
-              }
-              if (!fgets(buffer,LINE_SIZE,infile))
-                error_exit("Error reading symmetry operations");
-              cptr=buffer;
-            } while (j<3);
+	    if (qe_read_xml_floats(buffer,"<fractional_translation",
+				   symtr+3*i,3,infile)!=3)
+	      error_exit("Error reading sym translation");
           } /* end loop of sym ops */
         }
       }
@@ -517,6 +509,7 @@ void qe_xml_read(FILE* infile, char *filename, struct unit_cell *c,
         if(cptr){
           cptr+=6;
           e->energy=malloc(sizeof(double));
+	  if (!e->energy) error_exit("malloc error");
           i=sscanf(cptr,"%lf",e->energy);
           if (i!=1){
             fprintf(stderr,"Warning: error parsing energy\n");
@@ -553,12 +546,12 @@ void qe_xml_read(FILE* infile, char *filename, struct unit_cell *c,
     }
   }
 
-  if(!strstr(buffer,"</output>"))
+  if (!strstr(buffer,"</output>"))
     fprintf(stderr,"Warning, unexpected end to XML file\n");
 
   if (!c->basis) error_exit("No basis found in qe_xml_read");
 
-  if(nspec){
+  if (nspec){
     for(i=0;i<nspec;i++){
       pot=qe_pspot_file(species[i].p_pot,dict_get(m->dict,"QE_pseudo_dir"),cwd);
       if (pot){
@@ -572,7 +565,7 @@ void qe_xml_read(FILE* infile, char *filename, struct unit_cell *c,
 
   /* If timeseries, add final position */
 
-  if (ts){
+  if (ts->nsteps){
     if (ts->nc){
       ts->cells=realloc(ts->cells,(ts->nc+1)*sizeof(struct unit_cell));
       if (!ts->cells) error_exit("Realloc error for ts->cells");
@@ -649,8 +642,12 @@ void qe_xml_read(FILE* infile, char *filename, struct unit_cell *c,
     k_in->n=0;
   }
 
-  if (k_in->n) dict_add(m->dict,"QE_k_in",(void*)k_in);
-  
+  if (k_in->n)
+    dict_add(m->dict,"QE_k_in",(void*)k_in);
+  else{
+    free(k_in);
+    k_in=NULL;
+  }
 
   real2rec(c);
   addfrac(m->atoms,m->n,c->recip);
@@ -662,8 +659,16 @@ void qe_xml_read(FILE* infile, char *filename, struct unit_cell *c,
       addfrac(ts->m[i].atoms,ts->m[i].n,ts->cells[i].recip);
     }
   }
-  
+
+
   if (nspec){
+    vspin=0;
+    for(i=0;i<nspec;i++){
+      if ((species[i].mtheta!=0)||(species[i].mphi!=0)){
+	vspin=1;
+	break;
+      }
+    }
     for(i=0;i<m->n;i++){
       cptr=m->atoms[i].label;
       if (!cptr) cptr=atno2sym(m->atoms[i].atno);
@@ -671,7 +676,15 @@ void qe_xml_read(FILE* infile, char *filename, struct unit_cell *c,
 	for(j=0;j<nspec;j++){
 	  if (!strcasecmp(cptr,species[j].name)){
 	    m->atoms[i].chg=species[j].chg;
-	    m->atoms[i].spin=species[j].mag;
+	    if (vspin){
+	      m->atoms[i].vspin[0]=species[j].mag*sin(species[j].mtheta)*
+		cos(species[j].mphi);
+	      m->atoms[i].vspin[1]=species[j].mag*sin(species[j].mtheta)*
+		sin(species[j].mphi);
+	      m->atoms[i].vspin[2]=species[j].mag*cos(species[j].mtheta);
+	    }
+	    else
+	      m->atoms[i].spin=species[j].mag;
 	    break;
 	  }
 	}
@@ -693,6 +706,12 @@ void qe_xml_read(FILE* infile, char *filename, struct unit_cell *c,
       tmp=j;
     }
     dict_add(m->dict,"QE_atomic_species",cptr);
+    for(i=0;i<nspec;i++){
+      if (species[i].name) free(species[i].name);
+      if (species[i].p_pot) free(species[i].p_pot);
+    }
+    free(species);
+    species=NULL;
   }
 
 
@@ -708,12 +727,24 @@ void qe_xml_read(FILE* infile, char *filename, struct unit_cell *c,
 
       mat_f2a(mat,s->ops[i].mat,c->basis,c->recip);
 
-      for(j=0;j<3;j++){
-        s->ops[i].tr[j]=0;
-        for(jj=0;jj<3;jj++)
-          s->ops[i].tr[j]+=symtr[3*i+jj]*c->basis[jj][j];
+      /* The translations seem to differ in sign from other conventions */
+      s->ops[i].tr=NULL;
+      if ((symtr[3*i]!=0)||(symtr[3*i+1]!=0)||(symtr[3*i+2]!=0)){
+	s->ops[i].tr=malloc(3*sizeof(double));
+	if (!s->ops[i].tr)
+	  error_exit("Malloc error for symmetry translation");
+	for(j=0;j<3;j++){
+	  s->ops[i].tr[j]=0;
+	  for(jj=0;jj<3;jj++)
+	    s->ops[i].tr[j]-=symtr[3*i+jj]*c->basis[jj][j];
+	  if (aeq(s->ops[i].tr[j],-0.5)) s->ops[i].tr[j]=0.5;
+	}
       }
     }
+    free(symtr);
+    symtr=NULL;
+    free(symrel);
+    symrel=NULL;
   }
 
   if (e->eval) /* We store these in eV */
@@ -737,23 +768,28 @@ void qe_xml_read(FILE* infile, char *filename, struct unit_cell *c,
     while((cptr2>filename)&&(*cptr2!='/')) cptr2--;
     if (*cptr2=='/') i=cptr2-filename+1;
     den_name=malloc(1);
+    if (!den_name) error_exit("malloc error");
     *den_name=0;
     if (cptr2!=filename){
       den_name=realloc(den_name,strlen(filename)+1);
+      if (!den_name) error_exit("realloc error");
       strncpy(den_name,filename,i);
       den_name[i]=0;
     }
     den_name=realloc(den_name,strlen(den_name)+strlen("charge-density.dat")+1);
+    if (!den_name) error_exit("realloc error");
     strcat(den_name,"charge-density.dat");
     ch=fopen(den_name,"r");
     if (!ch){
       den_name2=NULL;
       if (!strcmp(filename+strlen(filename)-4,".xml")){
 	den_name2=malloc(strlen(filename)-3);
+	if (!den_name2) error_exit("malloc error");
 	strncpy(den_name2,filename,strlen(filename)-4);
 	den_name2[strlen(filename)-4]=0;
 	den_name2=realloc(den_name2,strlen(den_name2)+
 			  strlen(".save/charge-density.dat")+1);
+	if (!den_name2) error_exit("realloc error");
 	strcat(den_name2,".save/charge-density.dat");
 	ch=fopen(den_name2,"r");
       }
@@ -935,13 +971,18 @@ static void read_step(FILE *infile, char *buffer,struct time_series *ts){
   struct contents *m;
   
   ptr=strstr(buffer,"<step n_step=");
-  if (!ptr) return;
-
-  ptr+=strlen("<step n_step=")+1;
-  i=sscanf(ptr,"%d",&n);
-  if (n!=ts->nsteps+1){
-    fprintf(stderr,"Ignoring step %d as %d expected\n",n,ts->nsteps+1);
-    return;
+  if (ptr){ /* QE used to number steps */
+    ptr+=strlen("<step n_step=")+1;
+    i=sscanf(ptr,"%d",&n);
+    if (n!=ts->nsteps+1){
+      fprintf(stderr,"Ignoring step %d as %d expected\n",n,ts->nsteps+1);
+      return;
+    }
+  }
+  else{
+    ptr=strstr(buffer,"<step>");
+    if (!ptr) return;
+    n=ts->nsteps+1;
   }
 
   ts->nsteps=n;
@@ -997,18 +1038,17 @@ static void read_basis(FILE *infile, struct unit_cell *c){
   char *cptr,key[6];
   char buffer[LINE_SIZE+1];
   
-  c->basis=malloc(72);
+  c->basis=malloc(9*sizeof(double));
   if (!c->basis) error_exit("Malloc error for basis");
   for(i=0;i<3;i++){
-    sprintf(key,"<a%1d>",i+1);
+    sprintf(key,"<a%1d",i+1);
     fgets(buffer,LINE_SIZE,infile);
     cptr=strstr(buffer,key);
     if(!cptr){
       fprintf(stderr,"Failed to find %s\n",key);
       exit(1);
     }
-    j=sscanf(cptr+4,"%lf %lf %lf",c->basis[i],c->basis[i]+1,c->basis[i]+2);
-    if (j!=3){
+    if (qe_read_xml_floats(buffer,key,c->basis[i],3,infile)!=3){
       fprintf(stderr,"Failed to scan %s\n",key);
       exit(1);
     }
@@ -1020,7 +1060,7 @@ static void read_basis(FILE *infile, struct unit_cell *c){
 
 static void read_atoms(FILE *infile, struct contents *m){
   int i;
-  char *cptr,*cptr2,*cptr3;
+  char *cptr,*cptr2,*cptr3,ctmp;
   char buffer[LINE_SIZE+1];
   
   while(fgets(buffer,LINE_SIZE,infile)){
@@ -1046,19 +1086,87 @@ static void read_atoms(FILE *infile, struct contents *m){
       while(*cptr3!='"') cptr3++; /* cptr3 points one past end */
       if (!*cptr3) error_exit("Atom misparse 2.5");
       m->atoms[m->n-1].label=malloc((cptr3-cptr)+1);
+      if (!m->atoms[m->n-1].label) error_exit("malloc error");
       strncpy(m->atoms[m->n-1].label,cptr,cptr3-cptr);
       m->atoms[m->n-1].label[cptr3-cptr]=0;
     }
+    ctmp=*cptr2;
     *cptr2=0;
     m->atoms[m->n-1].atno=atsym2no(cptr);
+    *cptr2=ctmp;
     cptr=cptr2+1;
     while((*cptr)&&(*cptr!='>')) cptr++;
     if (!*cptr) error_exit("Atom misparse 3");
     cptr++;
-    i=sscanf(cptr,"%lf %lf %lf",m->atoms[m->n-1].abs,
-		 m->atoms[m->n-1].abs+1,m->atoms[m->n-1].abs+2);
-    if (i!=3) error_exit("Atom misparse 4");
+    if (qe_read_xml_floats(buffer,"<atom",m->atoms[m->n-1].abs,3,infile)!=3)
+      error_exit("Atom misparse 4");
     for(i=0;i<3;i++) m->atoms[m->n-1].abs[i]*=BOHR;
   }
 
 }
+
+/* key should have the form "<tag" and will match "<tag>" and
+ * <tag foo=bar>
+ */
+
+static int qe_read_xml_floats(char *line, char *key, double *data,
+			    int n, FILE* infile){
+  int i,tmp;
+  char *ptr,*ptr2;
+
+  //  fprintf(stderr,"qe_read_xml_floats called, line=%s\n"
+  //  	  "                           key=%s\n",line,key);
+  i=0;
+  ptr=line;
+  while (i==0){
+    ptr=strstr(ptr,key);
+    if (!ptr) error_exit("key error in qe_read_xml_floats");
+    i=1;
+    ptr2=ptr+strlen(key);
+    if ((*(ptr2)!=' ')&&(*(ptr2)!='>')){
+      ptr++;
+      i=0;
+    }
+  }
+
+  ptr+=strlen(key);
+  while ((*ptr)&&(*ptr!='>')) ptr++;
+  if (*ptr!='>') error_exit("key error 2 in qe_read_xml_floats");
+  ptr++;
+
+  i=0;
+  while(i<n){
+    if (sscanf(ptr,"%lf%n",data+i,&tmp)==1){
+      ptr+=tmp;
+      i++;
+    }
+    else{
+      while(isspace(*ptr)) ptr++;
+      if (*ptr==0){
+	if (!fgets(line,LINE_SIZE,infile))
+	  error_exit("read error or EOF in qe_read_xml_floats");
+	ptr=line;
+      }
+      else if ((*ptr=='<')&&(*(ptr+1)=='/')&&(strstr(ptr+2,key+1)==ptr+2))
+	return i;
+      else
+	error_exit("unexpected data in qe_read_xml_floats");
+    }
+  }
+    
+  while(isspace(*ptr)) ptr++;
+  if (*ptr==0){
+    fgets(line,LINE_SIZE,infile);
+    ptr=line;
+    while(isspace(*ptr)) ptr++;
+  }
+
+  if ((*ptr!='<')&&(*(ptr+1)!='/'))
+    error_exit("closing tag error in qe_read_xml_floats");
+  ptr+=2;
+  ptr2=strstr(ptr,key+1);
+  if (ptr2!=ptr)
+    error_exit("closing tag mismatch in qe_read_xml_floats");
+
+  return n;
+}
