@@ -1,6 +1,6 @@
 /* Conversion to arbitrary, specified supercells, with grid interpolation */
 
-/* Copyright (c) 2007 MJ Rutter 
+/* Copyright (c) 2007, 2022 MJ Rutter 
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -18,30 +18,30 @@
 
 #include<stdio.h>
 #include<stdlib.h>
+#include<string.h>  /* memcpy() */
 #include<math.h>
 
 #include "c2xsf.h"
 
-void vcross(double a[3],double b[3],double c[3]);
-
-
 void to235(int *i);
 void grid_interp(struct unit_cell *c, struct grid *grid,int old_fft[3],
                     double old_basis[3][3],double old_recip[3][3]);
+void grid_fft_interp(struct unit_cell *c, struct grid *grid,int old_fft[3],
+                    double old_basis[3][3],double old_recip[3][3]);
 /* From basis.c */
 int is_rhs(double b[3][3]);
-/* From ksym.c */
-void sym_basis(struct symmetry *s, struct unit_cell *c);
 /* From sort_atoms.c */
 int atom_sort(const void *a, const void *b);
 
 int super(struct unit_cell *c, struct contents *mtf,
            double new_basis[3][3], struct kpts *kp, struct symmetry *s, 
-           struct grid *gptr, int sflags){
-  int i,j,k,l,m,na,at,old_fft[3],same,rhs,quiet,kpts_only;
+	  struct grid *gptr, int sflags, int fft_in[3]){
+  int i,j,k,l,m,na,at,same,rhs,quiet,kpts_only,hit,new_fft[3];
+  int n_new_tr,finished;
   double old_basis[3][3],old_recip[3][3],old_vol,dtmp;
-  double new_in_old[3][3],abc[6],vtmp[3];
-  struct atom *old_atoms;
+  double new_in_old[3][3],abc[6],vtmp[3],vtmp2[3];
+  double (*new_tr)[3];
+  struct atom *old_atoms,old_axes[3];
   int *atom_ctr;
   int old_natoms;
   int scan_min[3],scan_max[3];
@@ -52,6 +52,11 @@ int super(struct unit_cell *c, struct contents *mtf,
   quiet=sflags&2;
   kpts_only=sflags&4;
 
+  if (fft_in)
+    for(i=0;i<3;i++) new_fft[i]=fft_in[i];
+  else
+    for(i=0;i<3;i++) new_fft[i]=0;
+  
   /* Correct new basis if not rhs and rhs wanted */
   /* Correction is exchange of 2nd and 3rd vectors */
   if (rhs){
@@ -111,6 +116,9 @@ int super(struct unit_cell *c, struct contents *mtf,
     }
     else{
       if (debug>1) fprintf(stderr,"Discarding MP grid\n");
+      if (!kp->kpts){ /* If we already have listed kpts, keep them */
+	mp_gen(kp,c); /* else generate them from MP params */
+      }
       free(kp->mp);
       kp->mp=NULL;
     }
@@ -121,6 +129,8 @@ int super(struct unit_cell *c, struct contents *mtf,
     struct contents rmtf;
 
     rcell.basis=malloc(9*sizeof(double));
+    if (!rcell.basis) error_exit("Malloc error in super");
+
     for(i=0;i<3;i++)
       for(j=0;j<3;j++){
 	rcell.basis[i][j]=c->recip[i][j];
@@ -133,12 +143,13 @@ int super(struct unit_cell *c, struct contents *mtf,
     addabs(rmtf.atoms,rmtf.n,rcell.basis);
 
     new_cell.basis=malloc(9*sizeof(double));
+    if (!new_cell.basis) error_exit("Malloc error in super");
     for(i=0;i<3;i++)
       for(j=0;j<3;j++)
 	new_cell.basis[i][j]=new_basis[i][j];
     real2rec(&new_cell);
     
-    if (super(&rcell,&rmtf,new_cell.recip,NULL,NULL,NULL,2)==0){
+    if (super(&rcell,&rmtf,new_cell.recip,NULL,NULL,NULL,2,NULL)==0){
 
       if (debug>2){
         fprintf(stderr,"New kpoints:\n");
@@ -148,6 +159,7 @@ int super(struct unit_cell *c, struct contents *mtf,
       }
 
       kp->n=rmtf.n;
+      free(kp->kpts);
       kp->kpts=rmtf.atoms;
       for(i=0;i<rmtf.n;i++) kp->kpts[i].wt*=fabs(new_cell.vol/c->vol);
     }
@@ -163,9 +175,14 @@ int super(struct unit_cell *c, struct contents *mtf,
   }
 
   if (kpts_only) return 0;
-  
+
   /* Save old cell */
 
+  init_atoms(old_axes,3);
+  for(i=0;i<3;i++)
+    for(j=0;j<3;j++)
+      old_axes[i].abs[j]=c->basis[i][j];
+  
   for(i=0;i<3;i++){
     for(j=0;j<3;j++){
       old_basis[i][j]=c->basis[i][j];
@@ -187,6 +204,9 @@ int super(struct unit_cell *c, struct contents *mtf,
 
   c->vol=fabs(c->vol);
 
+  addfrac(old_axes,3,c->recip);
+
+  if (debug) print_old_in_new(old_basis,c->basis);
 
   if (!quiet){
     if (debug) fprintf(stderr,"New cell volume %f (%g times old)\n",
@@ -208,8 +228,148 @@ int super(struct unit_cell *c, struct contents *mtf,
 
   /* Deal with symmetry operations */
 
-  if (s) sym_basis(s,c);
+  /* Do we have extra lattice vectors we wish to add to the sym ops? */
+  if ((s)&&(s->n)&&(dict_get(mtf->dict,"sym_expand"))){
+    n_new_tr=0;
+    new_tr=NULL;
+    for(i=0;i<3;i++){
+      for(j=0;j<3;j++){
+	old_axes[i].frac[j]=fmod(old_axes[i].frac[j],1.0);
+	if (old_axes[i].frac[j]<0) old_axes[i].frac[j]+=1;
+	if (aeq(old_axes[i].frac[j],1.0)) old_axes[i].frac[j]=0;
+      }
+      if (vmod2(old_axes[i].frac)>tol){
+	hit=0;
+	for(j=0;j<n_new_tr;j++){
+	  if ((dist(old_axes[i].frac[0],new_tr[j][0])<tol)&&
+	      (dist(old_axes[i].frac[1],new_tr[j][1])<tol)&&
+	      (dist(old_axes[i].frac[2],new_tr[j][2])<tol)){
+	    hit=1;
+	    break;
+	  }
+	}
+	if (!hit){
+	  new_tr=realloc(new_tr,(n_new_tr+1)*3*sizeof(double));
+	  if (!new_tr) error_exit("realloc error for new_tr");
+	  for(j=0;j<3;j++)
+	    new_tr[n_new_tr][j]=old_axes[i].frac[j];
+	  n_new_tr++;
+	}
+      }
+    }
+    if ((n_new_tr)&&(debug>1)){
+      fprintf(stderr,"Initial potential lattice vectors:\n");
+      for(i=0;i<n_new_tr;i++)
+	fprintf(stderr,"(%f,%f,%f)\n",new_tr[i][0],
+		new_tr[i][1],new_tr[i][2]);
+    }
+    finished=0;
+    while(!finished){
+      finished=1;
+      for(i=0;i<n_new_tr;i++){
+	for(j=i;j<n_new_tr;j++){
+	  for(k=0;k<3;k++){
+	    vtmp[k]=new_tr[i][k]+new_tr[j][k];
+	    vtmp[k]=fmod(vtmp[k],1.0);
+	    if (vtmp[k]<0) vtmp[k]+=1;
+	    if (aeq(vtmp[k],1.0)) vtmp[k]=0;
+	  }
+	  if (vmod2(vtmp)>tol){
+	    hit=0;
+	    for(k=0;k<n_new_tr;k++){
+	      if ((dist(vtmp[0],new_tr[k][0])<tol)&&
+		  (dist(vtmp[1],new_tr[k][1])<tol)&&
+		  (dist(vtmp[2],new_tr[k][2])<tol)){
+		hit=1;
+		break;
+	      }
+	    }
+	    if (!hit){
+	      new_tr=realloc(new_tr,(n_new_tr+1)*3*sizeof(double));
+	      if (!new_tr) error_exit("realloc error for new_tr");
+	      for(j=0;j<3;j++)
+		new_tr[n_new_tr][j]=vtmp[j];
+	      n_new_tr++;
+	      if (debug>1)
+		fprintf(stderr,"Adding (%f,%f,%f)\n",vtmp[0],vtmp[1],vtmp[2]);
+	      if (n_new_tr<20) finished=0;
+	      else error_exit("Too many new lattice vecs expanding symops");
+	    }
+	  }
+	}
+      }
+    }
+    
+    if ((n_new_tr)&&(debug>=1)){
+      fprintf(stderr,"Potential lattice vectors:\n");
+      for(i=0;i<n_new_tr;i++)
+	fprintf(stderr,"(%f,%f,%f)\n",new_tr[i][0],
+		new_tr[i][1],new_tr[i][2]);
+    }
+    
+    for(i=0;i<n_new_tr;i++){
+      for(j=0;j<3;j++)
+	vtmp[j]=new_tr[i][j];
+      hit=0;
+      for(j=0;j<s->n;j++){
+	if (!is_identity(s->ops[j].mat)) continue;
+	for(k=0;k<3;k++)
+	  vtmp2[k]=s->ops[j].tr[0]*c->recip[k][0]+
+	    s->ops[j].tr[1]*c->recip[k][1]+
+	    s->ops[j].tr[2]*c->recip[k][2];
+	if (debug>1)
+	  fprintf(stderr,"Comparing (%f,%f,%f) and (%f,%f,%f)\n",
+		  vtmp[0],vtmp[1],vtmp[2],
+		  vtmp2[0],vtmp2[1],vtmp2[2]);
+	if ((aeq(dist(vtmp[0],vtmp2[0]),0.0))&&
+	    (aeq(dist(vtmp[1],vtmp2[1]),0.0))&&
+	    (aeq(dist(vtmp[2],vtmp2[2]),0.0))){
+	  if (debug>1) fprintf(stderr,"Equal\n");
+	  hit=1;
+	  break;
+	}
+      }
+      if (hit) continue;
+      fprintf(stderr,"Trying to add translation of (%f,%f,%f) to sym ops\n",
+	      vtmp[0],vtmp[1],vtmp[2]);
+      if (debug) fprintf(stderr,"Starting with %d sym ops\n",s->n);
+      /* vtmp is in fractional coords, make vtmp2 in abs coords */
+      for(j=0;j<3;j++)
+	vtmp2[j]=new_tr[i][0]*c->basis[0][j]+
+	  new_tr[i][1]*c->basis[1][j]+
+	  new_tr[i][2]*c->basis[2][j];
 
+      s->ops=realloc(s->ops,2*s->n*sizeof(struct sym_op));
+      if (!s->ops){
+	fprintf(stderr,"realloc failed for %ld bytes\n",2*s->n*sizeof(struct sym_op));
+	exit(1);
+      }
+      //      if (!s->ops) error_exit("sym op realloc error in super.c");
+      for(j=0;j<s->n;j++){
+	memcpy(s->ops[s->n+j].mat,s->ops[j].mat,9*sizeof(double));
+	s->ops[s->n+j].tr=malloc(3*sizeof(double));
+	if (!s->ops[s->n+j].tr) error_exit("malloc error in super.c");
+	if (s->ops[j].tr)
+	  for(k=0;k<3;k++) s->ops[s->n+j].tr[k]=
+			     s->ops[j].tr[k]+vtmp2[k];
+	else
+	  for(k=0;k<3;k++) s->ops[s->n+j].tr[k]=vtmp2[k];
+      }
+      s->n*=2;
+
+      sym_basis(s,c);
+      if (debug) fprintf(stderr,"Ending with %d sym ops\n",s->n);
+    }
+
+  } /* if ((s)&&(dict_get(mtf->dict,"sym_expand"))) */
+
+  /* Remove excess sym ops */
+
+  if ((s)&&(s->n)){
+    sym_basis(s,c);
+    sym_group(s,c);
+  }
+  
   /* Worry about atoms */
 
   mtf->n=old_natoms*c->vol/old_vol+0.5;
@@ -353,27 +513,27 @@ int super(struct unit_cell *c, struct contents *mtf,
     fft_res=0;
     /* Find maximal grid point density */
     for(i=0;i<3;i++){
-      dtmp=gptr->size[i]/sqrt(old_basis[i][0]*old_basis[i][0]+
-                              old_basis[i][1]*old_basis[i][1]+
-                              old_basis[i][2]*old_basis[i][2]);
+      dtmp=gptr->size[i]/sqrt(vmod2(old_basis[i]));
       if (dtmp>fft_res) fft_res=dtmp;
     }
-    cart2abc(c,NULL,abc,NULL,0);
-    for(i=0;i<3;i++){
-      old_fft[i]=gptr->size[i];
-      gptr->size[i]=abc[i]*fft_res;
-      /* But if new axis simply a multiple of old, make new grid size
-       * that multiple of old too */
-      vcross(old_basis[i],c->basis[i],vtmp);
-      if (vmod2(vtmp)<1e-6*vmod2(old_basis[i])){
-        dtmp=sqrt(vmod2(c->basis[i])/vmod2(old_basis[i]));
-        gptr->size[i]=old_fft[i]*dtmp+0.5;
+    cart2abc(c,NULL,abc,NULL);
+    if ((new_fft[0]==0)||(new_fft[1]==0)||(new_fft[2]==0)){
+      for(i=0;i<3;i++){
+	new_fft[i]=abc[i]*fft_res;
+	/* But if new axis simply a multiple of old, make new grid size
+	 * that multiple of old too */
+	vcross(old_basis[i],c->basis[i],vtmp);
+	if (vmod2(vtmp)<1e-6*vmod2(old_basis[i])){
+	  dtmp=sqrt(vmod2(c->basis[i])/vmod2(old_basis[i]));
+	  new_fft[i]=gptr->size[i]*dtmp+0.5;
+	}
+	to235(new_fft+i);
       }
-      to235(gptr->size+i);
     }
     if (debug) fprintf(stderr,"New FFT grid is %d %d %d\n",
-                       gptr->size[0],gptr->size[1],gptr->size[2]);
-    grid_interp(c,gptr,old_fft,old_basis,old_recip);
+                       new_fft[0],new_fft[1],new_fft[2]);
+    grid_interp(c,gptr,new_fft,old_basis,old_recip);
+    if (debug) print_grid(gptr,c);
     gptr=gptr->next;
   }
   return 0;
@@ -395,27 +555,33 @@ void to235(int *i){
   to235(i);
 }
 
-void grid_interp(struct unit_cell *c, struct grid *grid, int old_fft[3],
+void grid_interp(struct unit_cell *c, struct grid *grid, int new_fft[3],
                     double old_basis[3][3], double old_recip[3][3]){
-  int i,j,k,l,ii,jj,kk,pt[3][2];
-  double pabs[3],pfrac[3],ifrac[3][2],*gnew;
-  double di,dj,dk,x,sum;
+  int i,j,k,l,ic,npts;
+  double pabs[3],pfrac[3],*gnew;
+  double di,dj,dk;
 
-  sum=0;
+  if (flags&FFT){
+    grid_fft_interp(c,grid,new_fft,old_basis,old_recip);
+    return;
+  }
 
+  if (!grid->data) return;
+  
   if (debug>1) fprintf(stderr,"Moving from %dx%dx%d grid to %dx%dx%d\n",
-		       old_fft[0],old_fft[1],old_fft[2],
-		       grid->size[0],grid->size[1],grid->size[2]);
+		       grid->size[0],grid->size[1],grid->size[2],
+		       new_fft[0],new_fft[1],new_fft[2]);
 
-  if (!(gnew=malloc(sizeof(double)*grid->size[0]*grid->size[1]*grid->size[2])))
+  npts=new_fft[0]*new_fft[1]*new_fft[2];
+  if (!(gnew=malloc(npts*grid->comps*sizeof(double))))
      error_exit("Malloc error in grid_interp");
 
-  for(i=0;i<grid->size[0];i++){
-    di=(double)i/grid->size[0];
-    for(j=0;j<grid->size[1];j++){
-      dj=(double)j/grid->size[1];
-      for(k=0;k<grid->size[2];k++){
-        dk=(double)k/grid->size[2];
+  for(i=0;i<new_fft[0];i++){
+    di=(double)i/new_fft[0];
+    for(j=0;j<new_fft[1];j++){
+      dj=(double)j/new_fft[1];
+      for(k=0;k<new_fft[2];k++){
+        dk=(double)k/new_fft[2];
         for(l=0;l<3;l++)
           pabs[l]=di*c->basis[0][l]+dj*c->basis[1][l]+dk*c->basis[2][l];
         /* convert to old cell.basis */
@@ -425,42 +591,24 @@ void grid_interp(struct unit_cell *c, struct grid *grid, int old_fft[3],
         /* reduce to unit cell */
           pfrac[l]=fmod(pfrac[l],1.0);
 	  if (pfrac[l]<0.0) pfrac[l]+=1.0;
-        /* convert to old grid */
-          pfrac[l]=pfrac[l]*old_fft[l];
-        /* and to int and fractional parts */
-          pt[l][0]=pfrac[l];
-          ifrac[l][1]=pfrac[l]-pt[l][0];
-	  if (pt[l][0]==old_fft[l]) pt[l][0]=0;
-	  if (pt[l][0]>old_fft[l]) error_exit("Impossible in grid_interp");
-        /* and the other side of the grid "cube" */
-          pt[l][1]=(pt[l][0]+1)%old_fft[l];
-          ifrac[l][0]=1-ifrac[l][1];
-        }
-        /* Trilinear interpolation */
-#define OX(x,y,z) grid->data[((x)*old_fft[1]+(y))*old_fft[2]+(z)]
-        x=0;
-        for(ii=0;ii<2;ii++)
-          for(jj=0;jj<2;jj++)
-            for(kk=0;kk<2;kk++)
-              x+=ifrac[0][ii]*ifrac[1][jj]*ifrac[2][kk]*
-                   OX(pt[0][ii],pt[1][jj],pt[2][kk]);
-        gnew[((i*grid->size[1])+j)*grid->size[2]+k]=x;
-	if (fabs(x)>99) {
-	  fprintf(stderr,"%lf\n",x);
-	  for(ii=0;ii<3;ii++) fprintf(stderr,"ifrac[0][%d]=%lf\n",ii,ifrac[0][i]);
-	  for(ii=0;ii<3;ii++)
-	    fprintf(stderr,"pt[i][0]=%5d  pt[i][1]=%5d\n",pt[ii][0],pt[ii][1]);  
 	}
-        sum+=x;
+	for(ic=0;ic<grid->comps;ic++){
+	  gnew[((i*new_fft[1])+j)*new_fft[2]+k+ic*npts]=
+	    interpolate0d(grid,pfrac,ic+1);
+	}
       }
     }
   }
 
   free(grid->data);
   grid->data=gnew;
+  for(i=0;i<3;i++)
+    grid->size[i]=new_fft[i];
 
-  if (debug>1) fprintf(stderr,"On new grid sum=%g int=%g\n",sum,
-                     sum*c->vol/(grid->size[0]*grid->size[1]*grid->size[2]));
+  if (debug>1){
+    fprintf(stderr,"On new grid\n");
+    print_grid(grid,c);
+  }
 }
 
 /* Independent code for the simple case of new vectors being integer
@@ -472,8 +620,8 @@ void simple_super(struct unit_cell *c, struct contents *m,
            struct grid *gptr){
   double new_basis[3][3],old_basis[3][3],*dptr;
   int new_fft[3],old_fft[3];
-  int i,j,k,p,q,old_n;
-  int ii,jj,kk;
+  int i,j,k,p,q,old_n,npts;
+  int ii,jj,kk,ic;
   int old_off,new_off;
   struct atom *old_atoms;
   
@@ -503,7 +651,7 @@ void simple_super(struct unit_cell *c, struct contents *m,
     print_basis(new_basis);
   }
   
-  super(c,m,new_basis,kp,s,gptr,4);
+  super(c,m,new_basis,kp,s,gptr,4,NULL);
 
   /* New basis */
 
@@ -515,7 +663,10 @@ void simple_super(struct unit_cell *c, struct contents *m,
   
   /* Deal with symmetry operations */
 
-  if (s) sym_basis(s,c);
+  if ((s)&&(s->n)){
+    sym_basis(s,c);
+    sym_group(s,c);
+  }
 
   /* Atoms */
   
@@ -552,17 +703,20 @@ void simple_super(struct unit_cell *c, struct contents *m,
       old_fft[i]=gptr->size[i];
     for(i=0;i<3;i++)
       new_fft[i]=expand[i]*old_fft[i];
-    dptr=malloc(new_fft[0]*new_fft[1]*new_fft[2]*sizeof(double));
+    npts=new_fft[0]*new_fft[1]*new_fft[2];
+    dptr=malloc(npts*gptr->comps*sizeof(double));
     if (!dptr) error_exit("Malloc error for new grid");
-    for(i=0;i<new_fft[0];i++){
-      ii=i%old_fft[0];
-      for(j=0;j<new_fft[1];j++){
-	jj=j%old_fft[1];
-	for(k=0;k<new_fft[2];k++){
-	  kk=k%old_fft[2];
-	  old_off=kk+old_fft[2]*(jj+ii*old_fft[1]);
-	  new_off=k+new_fft[2]*(j+i*new_fft[1]);
-	  dptr[new_off]=gptr->data[old_off];
+    for(ic=0;ic<gptr->comps;ic++){
+      for(i=0;i<new_fft[0];i++){
+	ii=i%old_fft[0];
+	for(j=0;j<new_fft[1];j++){
+	  jj=j%old_fft[1];
+	  for(k=0;k<new_fft[2];k++){
+	    kk=k%old_fft[2];
+	    old_off=kk+old_fft[2]*(jj+ii*old_fft[1])+ic*npts;
+	    new_off=k+new_fft[2]*(j+i*new_fft[1])+ic*npts;
+	    dptr[new_off]=gptr->data[old_off];
+	  }
 	}
       }
     }
@@ -573,3 +727,127 @@ void simple_super(struct unit_cell *c, struct contents *m,
     gptr=gptr->next;
   }
 }
+
+#define anint(x) ceil((x)-0.5)
+void grid_fft_interp(struct unit_cell *c, struct grid *grid, int new_fft[3],
+                    double old_basis[3][3], double old_recip[3][3]){
+  int i,j,k,m,ic,npts,npts_new,nn[3];
+  int hit,miss_nint,miss_out,ix,iy,iz,ii,jj,kk,okay;
+  double *rgrid,*rgrid_new,*g_new,*ptr,*ptr2,*pt,scale;
+  struct atom a;
+
+  if (!grid->data) return;
+
+  if (debug>1) fprintf(stderr,"Moving from %dx%dx%d to %dx%dx%d by FFT\n",
+		       grid->size[0],grid->size[1],grid->size[2],
+		       new_fft[0],new_fft[1],new_fft[2]);
+
+  npts_new=new_fft[0]*new_fft[1]*new_fft[2];
+  if (!(g_new=malloc(npts_new*grid->comps*sizeof(double))))
+     error_exit("Malloc error in grid_fft_interp");
+
+  npts=grid->size[0]*grid->size[1]*grid->size[2];
+  rgrid=malloc(npts*2*sizeof(double));
+  if (!rgrid)
+    error_exit("Malloc error in grid_fft_interp");
+
+  rgrid_new=malloc(npts_new*2*sizeof(double));
+  if (!rgrid)
+    error_exit("Malloc error in grid_fft_interp");
+
+  for(ic=0;ic<grid->comps;ic++){
+
+    ptr=grid->data+ic*npts;
+    ptr2=rgrid;
+    for(i=0;i<npts;i++){
+      *(ptr2++)=*(ptr++);
+      *(ptr2++)=0;
+    }
+    nn[0]=grid->size[2];
+    nn[1]=grid->size[1];
+    nn[2]=grid->size[0];
+    fft3d(rgrid,nn,-1);
+
+    for(i=0;i<2*npts_new;i++)
+      rgrid_new[i]=0;
+
+    hit=miss_nint=miss_out=0;
+    pt=a.frac;
+    for(i=0;i<new_fft[0];i++){
+      ii=i;
+      if (i>new_fft[0]/2) ii-=new_fft[0];
+      for(j=0;j<new_fft[1];j++){
+	jj=j;
+	if (j>new_fft[1]/2) jj-=new_fft[1];
+	for(k=0;k<new_fft[2];k++){
+	  kk=k;
+	  if (k>new_fft[2]/2) kk-=new_fft[2];
+	  pt[0]=ii;
+	  pt[1]=jj;
+	  pt[2]=kk;
+	  /* pt is in new reciprocal space. Where is it in old recip sp? */
+	  addabs(&a,1,c->recip);
+	  addfrac(&a,1,old_basis);
+	  okay=1;
+	  for(m=0;m<3;m++){
+	    if (!(aeq(pt[m],anint(pt[m])))){
+	      okay=0;
+	      miss_nint++;
+	      break;
+	    }
+	    if (anint(pt[m])>grid->size[m]/2){
+	      okay=0;
+	      miss_out++;
+	      break;
+	    }
+	    if (anint(pt[m])<-(grid->size[m]+1)/2){
+	      okay=0;
+	      miss_out++;
+	      break;
+	    }
+	  }
+	  
+	  if (okay){
+	    hit++;
+	    ix=anint(pt[0]);
+	    if (ix<0) ix+=grid->size[0];
+	    iy=anint(pt[1]);
+	    if (iy<0) iy+=grid->size[1];
+	    iz=anint(pt[2]);
+	    if (iz<0) iz+=grid->size[2];
+	    ptr=rgrid+2*(iz+grid->size[2]*(iy+grid->size[1]*ix));
+	    ptr2=rgrid_new+2*(k+new_fft[2]*(j+new_fft[1]*i));
+	    *ptr2=*ptr;
+	    *(ptr2+1)=*(ptr+1);
+	  }
+	}
+      }
+    }
+    nn[0]=new_fft[2];
+    nn[1]=new_fft[1];
+    nn[2]=new_fft[0];
+    fft3d(rgrid_new,nn,1);
+
+    scale=1./npts;
+    ptr=g_new+ic*npts_new;
+    ptr2=rgrid_new;
+    for(i=0;i<npts_new;i++){
+      *(ptr++)=*(ptr2)*scale;
+      ptr2+=2;
+    }
+
+    if (debug)
+      fprintf(stderr,"%d points moved, %d points lost (not gridable), "
+	      "%d points lost (outside)\n",
+	      hit,miss_nint,npts-hit-miss_out);
+    
+  }
+
+  free(rgrid_new);
+  free(rgrid);
+  free(grid->data);
+  grid->data=g_new;
+  for(i=0;i<3;i++)
+    grid->size[i]=new_fft[i];
+  
+}
