@@ -1,7 +1,7 @@
 /* Read a VASP CHGCAR or CHG, or also a POSCAR or a CONTCAR file
  */
 
-/* Copyright (c) 2017-2019 MJ Rutter 
+/* Copyright (c) 2017-2023 MJ Rutter 
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -59,8 +59,6 @@
 
 #define LINE_SIZE 2049
 
-void to235(int *i);
-
 static int formatted_read(double *dptr, unsigned int len, FILE *infile,
                           char *first);
 
@@ -74,7 +72,7 @@ void vasp_eigenval_read(FILE *infile, struct unit_cell *c, struct contents *m,
 void vasp_read(FILE* infile, char *filename, struct unit_cell *c,
                struct contents *m, struct kpts *kp, struct grid *gptr,
                struct es *elect){
-  int i,j,k,nspec,natoms,frac,fft[3],ffts[3],isize,locpot;
+  int i,j,k,nspec,natoms,frac,fft[3],ffts[3],isize,locpot,ns;
   int *nionsp,*species;
   char buffer[LINE_SIZE+1], *ptr,*p2;
   char *names[]={"POSCAR","CHGCAR","CONTCAR","CHG",NULL};
@@ -126,7 +124,7 @@ void vasp_read(FILE* infile, char *filename, struct unit_cell *c,
     
   /* Axes */
 
-  if (!(c->basis=malloc(72)))
+  if (!(c->basis=malloc(9*sizeof(double))))
     error_exit("Malloc error in vasp_read for c->basis");
   
   for(i=0;i<3;i++){
@@ -395,92 +393,130 @@ void vasp_read(FILE* infile, char *filename, struct unit_cell *c,
   
   if (!(flags&SPINDEN)) return;
   
-  if (!fgets(buffer,LINE_SIZE,infile)) return;
+  for(ns=0;ns<3;ns++){
 
-  /* We might have some augmentation occupancies.
-     For the moment, discard these. If norm-conserving pots used, this
-     section will be absent, as it also is in CHG rather than CHGCAR
-     files. */
-
-  while (!strncasecmp(buffer,"augmentation occupancies ",25)){
-    if (sscanf(buffer+25,"%d %d",&i,&j)!=2){
-      fprintf(stderr,"Failed to parse augmentation line. Ceasing reading\n");
-      return;
-    }
-    if (debug>2) fprintf(stderr,"Skipping augmentation occupancy line\n");
-    dptr=realloc(dptr,j*sizeof(double));
-    formatted_read(dptr,j,infile,NULL);
-    if (!fgets(buffer,LINE_SIZE,infile)){
-      fprintf(stderr,"Spin density not found\n");
-      return;
-    }
-  }
-
-  /* End of augmentation occupancies */
-
-  /* Might now have natom floats to read. These are the magnetic moments
-     for the atoms. But not if we have a CHG not CHGCAR file */
-
-  if (sscanf(buffer,"%d %d %d",ffts,ffts+1,ffts+2)!=3){
-    magmom=malloc(natoms*sizeof(double));
-    formatted_read(magmom,natoms,infile,buffer);
-    for(i=0;i<natoms;i++)
-      m->atoms[i].spin=magmom[i];
-    free(magmom);
-    if (!fgets(buffer,LINE_SIZE,infile)) return;
-  }
+    if (!fgets(buffer,LINE_SIZE,infile)) break;
   
-  if (sscanf(buffer,"%d %d %d",ffts,ffts+1,ffts+2)==3){
-    if ((ffts[0]==fft[0])&&
-        (ffts[1]==fft[1])&&
-        (ffts[2]==fft[2])){
-      if (debug) fprintf(stderr,"Spin density found\n");
-      gptr=grid_new(gptr);
+    /* We might have some augmentation occupancies.
+       For the moment, discard these. If norm-conserving pots used, this
+       section will be absent, as it also is in CHG rather than CHGCAR
+       files. */
 
-      isize=fft[0]*fft[1]*fft[2];
-      if(!(gptr->data=malloc(isize*sizeof(double))))
-        error_exit("Malloc error in vasp_read for grid data");
-      for(i=0;i<3;i++) gptr->size[i]=fft[i];
-
-      /* Data are not stored in the order that we want them */
-
-      if(!(dptr=malloc(isize*sizeof(double))))
-        error_exit("Malloc error in vasp_read for temp grid data");
-
-      if (formatted_read(dptr,isize,infile,NULL)) exit(1);
-
-      /* Now reverse order, writing consecutively for efficiency */
-
-      dptr2=gptr->data;
-      for(i=0;i<fft[0];i++)
-        for(j=0;j<fft[1];j++)
-          for(k=0;k<fft[2];k++)
-            *(dptr2++)=dptr[i+j*fft[0]+k*fft[0]*fft[1]];
-
-      free(dptr);
-      dptr=NULL;
-  
-      /* Rescale density */
-  
-      scale=1/fabs(c->vol);
-      if (flags&AU) scale*=BOHR*BOHR*BOHR;
-        
-      if(!(flags&RAW)){
-        for(i=0;i<isize;i++) gptr->data[i]*=scale;
-        gptr->name="Spin";
-        if (flags&AU)
-          add_cmt(m->comment,"Spin in e Bohr^-3");
-        else
-          add_cmt(m->comment,"Spin in e A^-3");
+    while (!strncasecmp(buffer,"augmentation occupancies ",25)){
+      k=25;
+      if (!strncasecmp(buffer+k,"(imaginary part) ",17))
+	k+=17;
+      if (sscanf(buffer+k,"%d %d",&i,&j)!=2){
+	fprintf(stderr,"Failed to parse augmentation line. Ceasing reading\n");
+	return;
+      }
+      if (debug>2) fprintf(stderr,"Skipping augmentation occupancy line\n");
+      dptr=realloc(dptr,j*sizeof(double));
+      formatted_read(dptr,j,infile,NULL);
+      if (!fgets(buffer,LINE_SIZE,infile)){
+	fprintf(stderr,"Spin density not found\n");
+	return;
       }
-      else{
-        gptr->name="Spin_raw";
-      } 
-      elect->nspins=2;
     }
+
+    /* End of augmentation occupancies */
+
+    /* We might have a single line containing only spaces (CHG not CHGCAR) */
+    ptr=buffer;
+    while(isspace(*ptr)) ptr++;
+    if (*ptr==0) fgets(buffer,LINE_SIZE,infile);
+    
+    /* Might now have natom floats to read. These are the magnetic moments
+       for the atoms. But not if we have a CHG not CHGCAR file */
+
+    if (sscanf(buffer,"%d %d %d",ffts,ffts+1,ffts+2)!=3){
+      magmom=malloc(natoms*sizeof(double));
+      formatted_read(magmom,natoms,infile,buffer);
+      if (ns==0)
+	for(i=0;i<natoms;i++)
+	  m->atoms[i].spin=magmom[i];
+      else if (ns==1)
+	for(i=0;i<natoms;i++){
+	  m->atoms[i].vspin[0]=m->atoms[i].spin;
+	  m->atoms[i].spin=0;
+	  m->atoms[i].vspin[1]=magmom[i];
+	}
+      else if (ns==2)
+	for(i=0;i<natoms;i++)
+	  m->atoms[i].vspin[2]=magmom[i];
+      free(magmom);
+      if (!fgets(buffer,LINE_SIZE,infile)) break;
+    }
+
+    if (sscanf(buffer,"%d %d %d",ffts,ffts+1,ffts+2)==3){
+      if ((ffts[0]==fft[0])&&
+	  (ffts[1]==fft[1])&&
+	  (ffts[2]==fft[2])){
+	if (debug) fprintf(stderr,"Spin density component found\n");
+	if (ns==0) gptr=grid_new(gptr);
+
+	isize=fft[0]*fft[1]*fft[2];
+	if(!(gptr->data=realloc(gptr->data,isize*(ns+1)*sizeof(double))))
+	  error_exit("Malloc error in vasp_read for grid data");
+	gptr->comps=ns+1;
+	if (ns==0) for(i=0;i<3;i++) gptr->size[i]=fft[i];
+
+	/* Data are not stored in the order that we want them */
+
+	if(!(dptr=malloc(isize*sizeof(double))))
+	  error_exit("Malloc error in vasp_read for temp grid data");
+
+	if (formatted_read(dptr,isize,infile,NULL))
+	  error_exit("Error reading spin component");
+
+	/* Now reverse order, writing consecutively for efficiency */
+
+	dptr2=gptr->data+ns*isize;
+	for(i=0;i<fft[0];i++)
+	  for(j=0;j<fft[1];j++)
+	    for(k=0;k<fft[2];k++)
+	      *(dptr2++)=dptr[i+j*fft[0]+k*fft[0]*fft[1]];
+
+	free(dptr);
+	dptr=NULL;
+      }
+      else
+	break;
+    }
+  } /* end for ns spin component loop */
+
+  if ((ns==0)||(!gptr->data)){
+    fprintf(stderr,"No spin found!\n");
+    return;
   }
-  else
-    fprintf(stderr,"Spin density not found\n");
+
+  if ((ns!=1)&&(ns!=3)){
+    fprintf(stderr,"Surprise! Found %d spin components, expected 0, 1 or 3\n",
+	    ns);
+    exit(1);
+  }
+
+  /* Rescale spin density */
+  
+  scale=1/fabs(c->vol);
+  if (flags&AU) scale*=BOHR*BOHR*BOHR;
+        
+  if(!(flags&RAW)){
+    for(i=0;i<isize*gptr->comps;i++) gptr->data[i]*=scale;
+    if (gptr->comps==1)
+      gptr->name="Spin";
+    else
+      gptr->name="Vector spin";
+    if (flags&AU)
+      add_cmt(m->comment,"Spin in e Bohr^-3");
+    else
+      add_cmt(m->comment,"Spin in e A^-3");
+  }
+  else{
+    gptr->name="Spin_raw";
+  }
+  elect->nspins=2;
+  if (gptr->comps==3) elect->nspinors=2;
 }
 
 static int formatted_read(double *dptr,unsigned int len, FILE *infile,
@@ -588,9 +624,10 @@ int vasp_grid_fill(int *pwgrid, double recip[3][3], double kpt[3],
 void vasp_psi_read(FILE* infile, char *filename, struct unit_cell *c,
                    struct contents *m, struct kpts *k, struct grid *gptr,
                    struct es *elect, int *i_grid){
-  int i,j,ns,tmp,nb,ik;
-  int nspins,reclen,nkpts,nbands,nplwv,*pwgrid,fft[3];
+  int i,j,ns,tmp,nb,ik,ii;
+  int nspins,nkpts,nbands,nplwv,*pwgrid,fft[3];
   int gamma,wt_warn,khdr_len,okay;
+  long reclen;
   double junk,version,ecut,kpt[3];
   float *psi_float;
   double *psi,basis[3][3];
@@ -610,7 +647,7 @@ void vasp_psi_read(FILE* infile, char *filename, struct unit_cell *c,
   fread(&junk,8,1,infile);
   if (junk!=(int)junk) error_exit("Unable to read WAVECAR");
   reclen=junk;
-  if (debug>2) fprintf(stderr,"reclen=%d  (%d doubles)\n",reclen,reclen/8);
+  if (debug>2) fprintf(stderr,"reclen=%ld  (%ld doubles)\n",reclen,reclen/8);
   
   fread(&junk,8,1,infile);
   nspins=junk;
@@ -648,6 +685,9 @@ void vasp_psi_read(FILE* infile, char *filename, struct unit_cell *c,
   fread(&junk,8,1,infile);
   nbands=junk;
 
+  if (debug>1) fprintf(stderr,"nkpts=%d, nbands=%d, nspins=%d\n",
+                       nkpts,nbands,nspins);
+  
   tmp=(3*nbands+4)*8;
   if (tmp>reclen){
     if (version==53300){
@@ -751,6 +791,7 @@ void vasp_psi_read(FILE* infile, char *filename, struct unit_cell *c,
       else k->kpts[ik].wt=1/(double)nkpts;
       for(j=0;j<3;j++) k->kpts[ik].frac[j]=kpt[j];
       pwgrid=malloc(3*nplwv*sizeof(int));
+      if (!pwgrid) error_exit("malloc error for pwgrid");
       if ((flags&BANDREAD)&&
           (inrange(ik+1,elect->kpt_range))&&
           (inrange(ns,elect->spin_range))){
@@ -770,6 +811,13 @@ void vasp_psi_read(FILE* infile, char *filename, struct unit_cell *c,
             for(i=0;i<3;i++) k->kpts[ik].frac[i]=kpt[i]=0;
             if (debug) fprintf(stderr,"Gamma point storage\n");
           }
+          else if ((nplwv==2*tmp)&&(nspins==1)){
+            if (elect->nspinors==1){
+              elect->nspinors=2;
+              if (debug) fprintf(stderr,"Spinor wavefunction detected\n");
+            }
+            for(ii=0;ii<3*(nplwv/2);ii++) pwgrid[ii+3*(nplwv/2)]=pwgrid[ii];
+          }
           else{
             fprintf(stderr,"Unexpected number of plane wave components: "
                     "skipping kpt %d (%.6f,%.6f,%.6f)\n",ik+1,
@@ -791,12 +839,22 @@ void vasp_psi_read(FILE* infile, char *filename, struct unit_cell *c,
             if (tmp!=2*nplwv) error_exit("Short read for psi_float");
 
             psi=malloc(2*nplwv*sizeof(double));
+            if (!psi) error_exit("malloc error for psi");
             for(i=0;i<2*nplwv;i++)
               psi[i]=psi_float[i];
             free(psi_float);
-            
-            band_process(psi,fft,pwgrid,nplwv,gamma,c,
-                         &gptr,elect,k,m,ik,0,ns,nb,i_grid);
+
+	    if ((flags&GCOEFF)&&(elect->nspinors==2)){
+		band_process(psi,fft,pwgrid,nplwv,gamma,c,
+			     &gptr,elect,k,m,ik,-1,ns,nb,i_grid);
+	    }
+	    else{
+	      for(ii=0;ii<elect->nspinors;ii++){
+		tmp=nplwv/elect->nspinors;
+		band_process(psi+2*ii*tmp,fft,pwgrid,nplwv,gamma,c,
+			     &gptr,elect,k,m,ik,ii,ns,nb,i_grid);
+	      }
+	    }
 
             free(psi);
 
@@ -819,7 +877,7 @@ void vasp_psi_read(FILE* infile, char *filename, struct unit_cell *c,
   }
 
   if (debug>2)
-    fprintf(stderr,"Final offset %d  reclen=%d\n",(int)ftell(infile),reclen);
+    fprintf(stderr,"Final offset %d  reclen=%ld\n",(int)ftell(infile),reclen);
   
 }
 
@@ -882,7 +940,7 @@ static void vasp_kpoints_read(FILE *infile, struct kpts *k){
   if (!k->kpts) error_exit("Malloc error for kpts");
   init_atoms(k->kpts,k->n);
   for(i=0;i<k->n;i++){
-    fgets(buffer,LINE_SIZE,infile);
+    if (!fgets(buffer,LINE_SIZE,infile)) error_exit("read error");
     if (sscanf(buffer,"%lf %lf %lf %lf",k->kpts[i].frac,
                k->kpts[i].frac+1,k->kpts[i].frac+2,&k->kpts[i].wt)!=4){
       fprintf(stderr,"Parse error for kpt");
